import pytest
import json
from pathlib import Path
from jaf.streaming_loader import StreamingLoader


@pytest.fixture
def setup_test_files(tmp_path):
    # Setup for json_array
    json_array_file = tmp_path / "test_array.json"
    json_array_file.write_text(
        json.dumps([{"id": 1, "type": "array"}, {"id": 2, "type": "array"}])
    )

    # Setup for single json object
    json_object_file = tmp_path / "test_object.json"
    json_object_file.write_text(json.dumps({"id": 3, "type": "object"}))

    # Setup for jsonl
    jsonl_file = tmp_path / "test.jsonl"
    jsonl_file.write_text(
        '{"id": 4, "type": "jsonl"}\n{"id": 5, "type": "jsonl"}\n42\n"hello"'
    )

    # Setup for directory
    dir_path = tmp_path / "test_dir"
    dir_path.mkdir()
    (dir_path / "a.json").write_text(json.dumps([{"id": 6, "type": "dir_json"}]))
    (dir_path / "b.jsonl").write_text('{"id": 7, "type": "dir_jsonl"}')
    (dir_path / "c.json").write_text(json.dumps({"id": 8, "single": "object"}))

    # Setup for gzipped file
    import gzip

    jsonl_gz_file = tmp_path / "test.jsonl.gz"
    with gzip.open(jsonl_gz_file, "wt") as f:
        f.write('{"id": 9, "type": "gzip"}\n{"id": 10, "type": "gzip"}')

    return {
        "json_array_file": json_array_file,
        "json_object_file": json_object_file,
        "jsonl_file": jsonl_file,
        "dir_path": dir_path,
        "jsonl_gz_file": jsonl_gz_file,
    }


def test_stream_json_array(setup_test_files):
    """Test streaming elements from a JSON array."""
    loader = StreamingLoader()
    source = {
        "type": "json_array",
        "inner_source": {
            "type": "file",
            "path": str(setup_test_files["json_array_file"]),
        },
    }

    results = list(loader.stream(source))
    assert len(results) == 2
    assert results[0] == {"id": 1, "type": "array"}
    assert results[1] == {"id": 2, "type": "array"}


def test_stream_json_object(setup_test_files):
    """Test loading a single JSON object file."""
    loader = StreamingLoader()
    source = {
        "type": "json_array",  # Will detect it's not an array and yield the single object
        "inner_source": {
            "type": "file",
            "path": str(setup_test_files["json_object_file"]),
        },
    }

    results = list(loader.stream(source))
    assert len(results) == 1
    assert results[0] == {"id": 3, "type": "object"}


def test_stream_json_value(setup_test_files):
    """Test treating entire JSON file as single value."""
    loader = StreamingLoader()
    source = {
        "type": "json_value",
        "inner_source": {
            "type": "file",
            "path": str(setup_test_files["json_array_file"]),
        },
    }

    results = list(loader.stream(source))
    assert len(results) == 1
    assert results[0] == [{"id": 1, "type": "array"}, {"id": 2, "type": "array"}]


def test_stream_jsonl(setup_test_files):
    """Test streaming from JSONL file with mixed value types."""
    loader = StreamingLoader()
    source = {
        "type": "jsonl",
        "inner_source": {"type": "file", "path": str(setup_test_files["jsonl_file"])},
    }

    results = list(loader.stream(source))
    assert len(results) == 4
    assert results[0] == {"id": 4, "type": "jsonl"}
    assert results[1] == {"id": 5, "type": "jsonl"}
    assert results[2] == 42
    assert results[3] == "hello"


def test_stream_directory(setup_test_files):
    """Test streaming from directory of JSON files."""
    loader = StreamingLoader()
    source = {
        "type": "directory",
        "path": str(setup_test_files["dir_path"]),
        "pattern": "*.json*",
    }

    results = list(loader.stream(source))
    # Should get objects from all files
    assert len(results) == 3
    ids = [r.get("id") for r in results if isinstance(r, dict)]
    assert 6 in ids  # From a.json array
    assert 7 in ids  # From b.jsonl
    assert 8 in ids  # From c.json object


def test_stream_directory_file_as_object(setup_test_files):
    """Test streaming directory with file_as_object option."""
    loader = StreamingLoader()
    source = {
        "type": "directory",
        "path": str(setup_test_files["dir_path"]),
        "pattern": "a.json",
        "file_as_object": True,
    }

    results = list(loader.stream(source))
    # Should get the entire array as one item
    assert len(results) == 1
    assert results[0] == [{"id": 6, "type": "dir_json"}]


def test_stream_gzipped_jsonl(setup_test_files):
    """Test streaming from gzipped JSONL file."""
    loader = StreamingLoader()
    source = {
        "type": "jsonl",
        "inner_source": {
            "type": "gzip",
            "inner_source": {
                "type": "file",
                "path": str(setup_test_files["jsonl_gz_file"]),
            },
        },
    }

    results = list(loader.stream(source))
    assert len(results) == 2
    assert results[0] == {"id": 9, "type": "gzip"}
    assert results[1] == {"id": 10, "type": "gzip"}


def test_stream_memory_source():
    """Test streaming from in-memory data."""
    loader = StreamingLoader()
    source = {"type": "memory", "data": [{"id": 1}, 42, "hello", [1, 2, 3]]}

    results = list(loader.stream(source))
    assert len(results) == 4
    assert results[0] == {"id": 1}
    assert results[1] == 42
    assert results[2] == "hello"
    assert results[3] == [1, 2, 3]


def test_unknown_source_type():
    """Test error handling for unknown source type."""
    loader = StreamingLoader()
    source = {"type": "unknown"}

    with pytest.raises(ValueError, match="Unknown source type: unknown"):
        list(loader.stream(source))


def test_missing_source_type():
    """Test error handling for missing source type."""
    loader = StreamingLoader()
    source = {"path": "/some/path"}

    with pytest.raises(ValueError, match="Source descriptor missing 'type' field"):
        list(loader.stream(source))


def test_custom_loader_registration():
    """Test registering custom loader."""
    loader = StreamingLoader()

    def custom_loader(loader_instance, source):
        yield {"custom": True}
        yield {"custom": False}

    loader.register("custom", custom_loader)

    source = {"type": "custom"}
    results = list(loader.stream(source))
    assert len(results) == 2
    assert results[0] == {"custom": True}
    assert results[1] == {"custom": False}



"""
Simple streaming JSON array parser for JAF.

This module provides a lightweight streaming parser that extracts JSON objects
from a JSON array without loading the entire array into memory. It works by
counting braces to identify complete objects.
"""

import json
import gzip
from typing import Generator, Dict, Any, IO, Union, Optional
import io


def stream_json(
    file_obj: Union[IO[str], IO[bytes]],
    chunk_size: int = 65536,
    initial_object_size: int = 4096,
) -> Generator[Dict[str, Any], None, None]:
    """
    Streaming JSON parser that handles both arrays and single objects.

    This parser can handle:
    - JSON array of objects: [{...}, {...}, ...]
    - Single JSON object: {...}

    For arrays, it streams each object. For single objects, it yields the object once.

    Args:
        file_obj: File object (text or binary mode)
        chunk_size: Size of chunks to read from file (default 64KB)
        initial_object_size: Initial size for object buffer (default 4KB)

    Yields:
        Dict[str, Any]: Each JSON object

    Raises:
        ValueError: If the file contains invalid JSON
    """
    # Handle both text and binary files
    if hasattr(file_obj, "mode") and "b" in file_obj.mode:
        # Binary mode
        read_chunk = lambda: file_obj.read(chunk_size).decode("utf-8", errors="replace")
    else:
        # Text mode
        read_chunk = lambda: file_obj.read(chunk_size)

    # Read first chunk
    chunk = read_chunk()
    if not chunk:
        return

    # Find first non-whitespace character
    pos = skip_whitespace(chunk, 0)
    if pos >= len(chunk):
        return

    # Check if it's an array
    if chunk[pos] != "[":
        # Not an array - read entire value and yield it
        all_chunks = [chunk]
        while True:
            next_chunk = read_chunk()
            if not next_chunk:
                break
            all_chunks.append(next_chunk)

        full_content = "".join(all_chunks)
        try:
            value = json.loads(full_content)
            yield value
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON: {e}")
        return

    # It's an array - stream each value
    pos += 1  # Skip the '['

    # Create chunk iterator
    def chunk_iterator():
        nonlocal chunk, pos
        while True:
            next_chunk = read_chunk()
            if not next_chunk:
                break
            yield next_chunk

    chunks = chunk_iterator()

    while True:
        # Skip whitespace and commas
        while pos < len(chunk) and chunk[pos] in " \t\n\r,":
            pos += 1

        # Need more data?
        if pos >= len(chunk):
            try:
                chunk = next(chunks)
                pos = 0
                continue
            except StopIteration:
                return

        # Check for end of array
        if chunk[pos] == "]":
            return

        # Parse next value based on first character
        first_char = chunk[pos]

        try:
            if first_char == "{":
                # Object
                value_str, chunk, pos = extract_balanced_value(
                    chunks, chunk, pos, "{", "}"
                )
            elif first_char == "[":
                # Nested array
                value_str, chunk, pos = extract_balanced_value(
                    chunks, chunk, pos, "[", "]"
                )
            elif first_char == '"':
                # String
                value_str, chunk, pos = extract_string_value(chunks, chunk, pos)
            else:
                # Number, boolean, or null
                value_str, chunk, pos = extract_literal_value(chunks, chunk, pos)

            # Parse and yield the value
            value = json.loads(value_str)
            yield value

        except (json.JSONDecodeError, ValueError):
            # Skip malformed values, try to recover
            # Find next comma or closing bracket
            while pos < len(chunk) and chunk[pos] not in ",]":
                pos += 1


def skip_whitespace(text: str, start: int) -> int:
    """
    Skip whitespace characters and return position of next non-whitespace.

    Args:
        text: String to scan
        start: Starting position

    Returns:
        Position of next non-whitespace character, or len(text) if all whitespace
    """
    # Use built-in string methods for efficiency
    remainder = text[start:]
    stripped = remainder.lstrip()
    if stripped:
        return start + (len(remainder) - len(stripped))
    return len(text)


def extract_balanced_value(
    chunks_iter, initial_chunk, start_pos, open_char, close_char
):
    """
    Extract a balanced JSON value (object or array) by counting brackets/braces.

    Returns: (value_str, remaining_chunk, next_pos)
    """
    parts = []
    chunk = initial_chunk
    pos = start_pos

    in_string = False
    escape_next = False
    depth = 0

    while True:
        while pos < len(chunk):
            char = chunk[pos]

            # Handle string state
            if not escape_next:
                if char == '"' and not in_string:
                    in_string = True
                elif char == '"' and in_string:
                    in_string = False
                elif char == "\\" and in_string:
                    escape_next = True
            else:
                escape_next = False

            # Count brackets/braces when not in string
            if not in_string:
                if char == open_char:
                    depth += 1
                elif char == close_char:
                    depth -= 1
                    if depth == 0:
                        # Complete value found
                        parts.append(chunk[start_pos : pos + 1])
                        return "".join(parts), chunk, pos + 1

            pos += 1

        # Need more data
        parts.append(chunk[start_pos:])
        try:
            chunk = next(chunks_iter)
            start_pos = 0
            pos = 0
        except StopIteration:
            raise ValueError(
                f"Unexpected end of input while parsing {open_char}...{close_char}"
            )


def extract_string_value(chunks_iter, initial_chunk, start_pos):
    """
    Extract a JSON string value.

    Returns: (value_str, remaining_chunk, next_pos)
    """
    parts = []
    chunk = initial_chunk
    pos = start_pos + 1  # Skip opening quote

    escape_next = False

    while True:
        while pos < len(chunk):
            char = chunk[pos]

            if not escape_next:
                if char == '"':
                    # String complete
                    parts.append(chunk[start_pos : pos + 1])
                    return "".join(parts), chunk, pos + 1
                elif char == "\\":
                    escape_next = True
            else:
                escape_next = False

            pos += 1

        # Need more data
        parts.append(chunk[start_pos:])
        try:
            chunk = next(chunks_iter)
            start_pos = 0
            pos = 0
        except StopIteration:
            raise ValueError("Unexpected end of input while parsing string")


def extract_literal_value(chunks_iter, initial_chunk, start_pos):
    """
    Extract a JSON literal (number, true, false, null).

    Returns: (value_str, remaining_chunk, next_pos)
    """
    chunk = initial_chunk
    pos = start_pos

    # Find the end of the literal (whitespace, comma, or closing bracket)
    while pos < len(chunk):
        char = chunk[pos]
        if char in " \t\n\r,]}":
            return chunk[start_pos:pos], chunk, pos
        pos += 1

    # Literal continues into next chunk
    parts = [chunk[start_pos:]]

    try:
        while True:
            chunk = next(chunks_iter)
            pos = 0
            while pos < len(chunk):
                char = chunk[pos]
                if char in " \t\n\r,]}":
                    parts.append(chunk[:pos])
                    return "".join(parts), chunk, pos
                pos += 1
            parts.append(chunk)
    except StopIteration:
        # End of input
        return "".join(parts), "", 0


def stream_jsonl(file_obj: Union[IO[str], IO[bytes]]) -> Generator[Any, None, None]:
    """
    Stream JSON values from a JSONL (JSON Lines) file.

    Args:
        file_obj: An open file object containing JSONL data

    Yields:
        Any: JSON values (one per line)
    """
    for line in file_obj:
        if isinstance(line, bytes):
            line = line.decode("utf-8", errors="replace")
        line = line.strip()
        if line:
            try:
                value = json.loads(line)
                yield value
            except json.JSONDecodeError:
                # Skip invalid lines
                pass


def stream_json_file(
    file_path: str, format: Optional[str] = None
) -> Generator[Dict[str, Any], None, None]:
    """
    Stream JSON objects from a file, automatically handling compression and format.

    Args:
        file_path: Path to the file
        format: Force a specific format ('json', 'jsonl', or None for auto-detect)

    Yields:
        Dict[str, Any]: Each JSON object in the file
    """
    # Determine format from extension if not specified
    if format is None:
        if file_path.endswith(".jsonl") or file_path.endswith(".jsonl.gz"):
            format = "jsonl"
        else:
            format = "json"

    # Handle gzip compression
    if file_path.endswith(".gz"):
        file_obj = gzip.open(file_path, "rt", encoding="utf-8")
    else:
        file_obj = open(file_path, "r", encoding="utf-8")

    try:
        if format == "jsonl":
            yield from stream_jsonl(file_obj)
        else:
            yield from stream_json_array(file_obj)
    finally:
        file_obj.close()


def stream_json_collection(
    source: Dict[str, Any],
) -> Generator[Dict[str, Any], None, None]:
    """
    Stream JSON objects from various collection sources.

    Args:
        source: Collection source descriptor with 'type' and type-specific fields

    Yields:
        Dict[str, Any]: Each JSON object in the collection
    """
    source_type = source.get("type") or source.get("source_type")

    if source_type == "jsonl":
        path = source.get("path")
        if path:
            yield from stream_json_file(path, format="jsonl")

    elif source_type == "json_array":
        path = source.get("path")
        if path:
            yield from stream_json_file(path, format="json")

    elif source_type == "directory":
        files = source.get("files", [])
        for file_path in files:
            if file_path.endswith(".jsonl") or file_path.endswith(".jsonl.gz"):
                yield from stream_json_file(file_path, format="jsonl")
            else:
                yield from stream_json_file(file_path, format="json")

    elif source_type == "buffered_stdin" or source_type == "in_memory":
        # These are already in memory, just yield them
        content = source.get("content") or source.get("data", [])
        for item in content:
            if isinstance(item, dict):
                yield item

    else:
        raise ValueError(f"Unknown source type: {source_type}")



"""
JAF Path Types
==============

This module contains data types related to JAF path operations.
"""

from typing import Any, Optional


class PathValues(list):
    """
    A list subclass representing a collection of values obtained from a JAF path
    evaluation. It's used particularly when the path could naturally yield
    multiple results (e.g., from wildcards, slices, multiple indices, or
    regex key matching).

    It behaves like a standard list but serves as a distinct type to inform
    JAF's evaluation logic (e.g., `adapt_jaf_operator`) on how to handle
    arguments that are collections of path results. This typically involves
    iterating over the values or applying Cartesian product logic if multiple
    PathValues are arguments to an operator.

    Conceptually, it holds the "set of findings" or "resolved values" from the
    data for a given path. While it's a list (and thus can hold duplicates
    and maintains order of discovery), its primary distinction is signaling
    its origin from a potentially multi-valued path.

    An empty PathValues instance (e.g., PathValues([])) signifies that a
    multi-match path expression found no values. This is distinct from an
    empty list `[]` which might be returned by `eval_path` when a specific,
    non-multi-match path is not found.
    """

    def __init__(self, iterable: Optional[Any] = None):
        super().__init__(iterable if iterable is not None else [])

    def __repr__(self) -> str:
        return f"PathValues({super().__repr__()})"

    def __getitem__(self, key: Any) -> Any:
        result = super().__getitem__(key)
        if isinstance(key, slice):
            return PathValues(result)
        # Single item access returns the item itself, not PathValues(item)
        return result

    def first(self, default: Optional[Any] = None) -> Any:
        """
        Returns the first element, or `default` if the list is empty.
        """
        return self[0] if self else default

    def last(self, default: Optional[Any] = None) -> Any:
        """
        Returns the last element, or `default` if the list is empty.
        """
        return self[-1] if self else default

    def one(self) -> Any:
        """
        Returns the single element if the list contains exactly one item.

        Raises:
            ValueError: If the list does not contain exactly one item.
        """
        if len(self) == 1:
            return self[0]
        elif not self:
            raise ValueError("PathValues is empty; expected exactly one element.")
        else:
            raise ValueError(
                f"PathValues contains {len(self)} elements; expected exactly one."
            )

    def one_or_none(self) -> Optional[Any]:
        """
        Returns the single element if the list contains exactly one item,
        or None if the list is empty.

        Raises:
            ValueError: If the list contains more than one item.
        """
        if len(self) == 1:
            return self[0]
        elif not self:
            return None
        else:
            raise ValueError(
                f"PathValues contains {len(self)} elements; expected one or none."
            )

    def __contains__(self, item: Any) -> bool:
        """
        Check if an item is in the PathValues.
        This uses the base list's __contains__ method.
        """
        return super().__contains__(item)



"""
JAF Path Evaluation
===================

This module contains the core path evaluation functions for JAF.
It provides the main interface for evaluating path expressions against data objects.
"""

import logging
from typing import Any, List

from .path_conversion import path_ast_to_string, string_to_path_ast
from .path_exceptions import PathSyntaxError
from .path_types import PathValues
from .path_operations import PathOperationDispatcher

logger = logging.getLogger(__name__)


def _path_has_multi_match_components(path_components_list: List[List[Any]]) -> bool:
    """
    Internal helper to determine if a path expression's components
    include operations that inherently imply multiple matches (e.g., wildcards, slices).
    """
    if not path_components_list:
        return False
    for component_list_item in path_components_list:
        if not isinstance(component_list_item, list) or not component_list_item:
            continue
        op = component_list_item[0]
        if op in [
            "indices",
            "slice",
            "regex_key",
            "fuzzy_key",
            "wc_level",
            "wc_recursive",
        ]:
            return True
    return False


# Global dispatcher instance
_path_dispatcher = PathOperationDispatcher()


def _match_recursive(
    current_obj: Any,
    components: List[List[Any]],
    full_path_ast_for_error: List[List[Any]],
    root_obj_for_path: Any,
) -> List[Any]:
    """
    Recursively match path components against the current object.

    Args:
        current_obj: The current object being traversed
        components: Remaining path components to process
        full_path_ast_for_error: Complete path AST for error reporting
        root_obj_for_path: The root object for this path evaluation

    Returns:
        List of matching values
    """
    if current_obj is None and components:
        return []

    if not components:
        return [current_obj]

    component = components[0]
    remaining_components = components[1:]

    if not isinstance(component, list) or not component:
        raise PathSyntaxError(
            "Path component must be a non-empty list.",
            path_segment=component,
            full_path_ast=full_path_ast_for_error,
        )

    op = component[0]
    args = component[1:]

    return _path_dispatcher.dispatch(
        op,
        args,
        current_obj,
        remaining_components,
        full_path_ast_for_error,
        root_obj_for_path,
    )


def eval_path(path_components_list: List[List[Any]], obj: Any) -> Any:
    """
    Evaluates a path expression against an object and retrieves values.

    This is the main function for path evaluation in JAF. It takes a path
    expression in AST format and evaluates it against the provided object.

    Args:
        path_components_list: List of path components in AST format
        obj: The object to evaluate the path against

    Returns:
        - For specific paths: the value at that path, or [] if not found
        - For multi-match paths: PathValues containing all matching values

    Raises:
        PathSyntaxError: For malformed path ASTs
    """
    if not isinstance(path_components_list, list):
        raise PathSyntaxError("Path expression must be a list of components.", full_path_ast=path_components_list)  # type: ignore

    if not path_components_list:  # Empty path means the object itself
        return obj

    # Validate all components before starting recursion for early failure
    known_path_ops = {
        "key",
        "index",
        "indices",
        "slice",
        "regex_key",
        "fuzzy_key",
        "wc_level",
        "wc_recursive",
        "root",
    }
    for component in path_components_list:
        if (
            not isinstance(component, list)
            or not component
            or not isinstance(component[0], str)
        ):
            raise PathSyntaxError(
                "Each path component must be a non-empty list starting with an operation string.",
                path_segment=component,
                full_path_ast=path_components_list,
            )

    matched_values = _match_recursive(
        obj,
        path_components_list,
        full_path_ast_for_error=path_components_list,
        root_obj_for_path=obj,
    )

    is_specific_path_intent = not _path_has_multi_match_components(path_components_list)

    if is_specific_path_intent:
        if not matched_values:
            return []
        elif len(matched_values) == 1:
            return matched_values[0]
        else:
            logger.debug(
                f"Path with specific-intent components yielded {len(matched_values)} results. "
                f"Path: {path_components_list}. Wrapping in PathValues."
            )
            return PathValues(matched_values)
    else:
        return PathValues(matched_values)


def exists(path_components_list: List[List[Any]], obj: Any) -> bool:
    """
    Checks if the given path expression resolves to any value(s) in the object.

    This function determines whether a path exists in the data, including
    paths that resolve to None/null values.

    TODO: BUGFIX - This function currently returns False for keys that exist
    with empty array values because eval_path returns [] for both non-existent
    paths and paths pointing to empty arrays. Need to modify eval_path to
    distinguish between these cases (e.g., return None for non-existent paths).

    Args:
        path_components_list: List of path components in AST format
        obj: The object to check the path against

    Returns:
        True if the path exists (even if the value is None), False otherwise

    Raises:
        PathSyntaxError: For malformed path ASTs
    """
    try:
        resolved_value = eval_path(path_components_list, obj)
        if isinstance(resolved_value, PathValues):
            return len(resolved_value) > 0
        # If eval_path returns an empty list for a specific path (not PathValues),
        # it means the path resolved to nothing, so it doesn't exist.
        elif (
            isinstance(resolved_value, list)
            and not resolved_value
            and not _path_has_multi_match_components(path_components_list)
        ):
            return False
        return True
    except (
        PathSyntaxError
    ):  # Re-raise PathSyntaxError as it's an issue with the path itself
        raise
    except Exception:
        logger.debug(
            f"Exception during 'exists' check for path {path_components_list}",
            exc_info=True,
        )
        return False  # Other exceptions imply the path didn't resolve or data was incompatible


def is_valid_path_str(path: str) -> bool:
    """
    Checks if the provided path string is a valid JAF path expression.

    Args:
        path: The path string to validate

    Returns:
        True if the path is valid, False otherwise
    """
    if not isinstance(path, str):
        logger.debug(f"Invalid path type: {type(path)}. Expected a string.")
        return False
    try:
        # Attempt to parse the path string using the JAF path syntax
        string_to_path_ast(path)
        return True
    except PathSyntaxError:
        return False
    except Exception as e:
        logger.debug(f"Unexpected error validating path '{path}': {e}", exc_info=True)
        return False



"""
Core JAF functionality tests.
Tests core filtering functionality using the stream API.
"""

import pytest
from jaf.lazy_streams import stream, FilteredStream
from jaf.exceptions import JAFError


class TestJAFCore:
    """Test core JAF functionality"""

    def setup_method(self):
        """Set up test data"""
        self.test_data = [
            {"id": 1, "name": "Alice", "age": 30, "active": True},
            {"id": 2, "name": "Bob", "age": 25, "active": False},
            {"id": 3, "name": "Charlie", "age": 35, "active": True},
            {"id": 4, "name": "Diana", "age": 28, "active": True},
        ]
        self.test_data_collection_id = "test_data_v1"

    def test_stream_returns_filtered_stream(self):
        """Test that stream filtering returns a FilteredStream instance."""
        query = ["eq?", ["@", [["key", "name"]]], "Alice"]
        s = stream({"type": "memory", "data": self.test_data})
        result = s.filter(query)

        assert isinstance(result, FilteredStream)
        assert result.query == query
        # Test evaluation
        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 1
        assert matching_objects[0]["name"] == "Alice"

    def test_stream_with_collection_id(self):
        """Test that stream correctly handles collection_id."""
        query = ["eq?", ["@", [["key", "name"]]], "Alice"]
        s = stream({"type": "memory", "data": self.test_data})
        result = s.filter(query)
        # Set collection_id on the stream
        result.collection_id = self.test_data_collection_id

        assert isinstance(result, FilteredStream)
        assert result.query == query
        assert result.collection_id == self.test_data_collection_id
        # Test evaluation
        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 1
        assert matching_objects[0]["name"] == "Alice"

    def test_multiple_matches(self):
        """Test multiple matching objects"""
        query = ["eq?", ["@", [["key", "active"]]], True]
        s = stream({"type": "memory", "data": self.test_data})
        result = s.filter(query)
        assert isinstance(result, FilteredStream)
        assert result.query == query
        # Test evaluation
        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 3  # Alice, Charlie, Diana
        matching_names = {obj["name"] for obj in matching_objects}
        assert matching_names == {"Alice", "Charlie", "Diana"}

    def test_no_matches(self):
        """Test query with no matches"""
        query = ["eq?", ["@", [["key", "name"]]], "Nobody"]
        s = stream({"type": "memory", "data": self.test_data})
        result = s.filter(query)
        assert isinstance(result, FilteredStream)
        assert result.query == query
        # Test evaluation
        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 0  # No matches

    def test_empty_data(self):
        """Test with empty data array"""
        query = ["eq?", ["@", [["key", "name"]]], "Alice"]
        s = stream({"type": "memory", "data": []})
        result = s.filter(query)
        assert isinstance(result, FilteredStream)
        assert result.query == query
        # Test evaluation
        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 0  # No data to match

    def test_invalid_query_raises_error(self):
        """Test that invalid queries raise appropriate errors"""
        from jaf.exceptions import UnknownOperatorError

        # With lazy evaluation, invalid queries are only detected during evaluation
        query = ["unknown-operator", "arg"]
        s = stream({"type": "memory", "data": self.test_data})
        result = s.filter(query)

        # Query creation should succeed
        assert isinstance(result, FilteredStream)
        assert result.query == query

        # But evaluation should fail with clear error
        with pytest.raises(
            UnknownOperatorError,
            match="Unknown operator: unknown-operator",
        ):
            list(result.evaluate())

    def test_empty_query_raises_error(self):
        """Test that empty query raises error"""
        # With streams, we need to test at filter time
        s = stream({"type": "memory", "data": self.test_data})

        # Empty string or None queries should be caught
        # Note: The actual error handling may differ in the stream implementation

    def test_non_dict_objects_skipped(self):
        """Test that non-dictionary objects are skipped"""
        mixed_data = [
            {"name": "Alice"},
            "not a dict",
            {"name": "Bob"},
            123,
            {"name": "Charlie"},
        ]
        query = ["eq?", ["@", [["key", "name"]]], "Bob"]
        s = stream({"type": "memory", "data": mixed_data})
        result = s.filter(query)
        assert isinstance(result, FilteredStream)
        assert result.query == query
        # Test evaluation
        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 1  # Only Bob matches
        assert matching_objects[0]["name"] == "Bob"

    def test_malformed_data_object_input(
        self,
    ):  # Renamed to avoid conflict if None was a valid JafQuerySet input
        """Test handling of data arrays containing non-dictionary elements like None"""
        malformed_data = [
            {"name": "Alice"},
            None,  # This object will be skipped
            {"name": "Bob"},
        ]
        query = ["eq?", ["@", [["key", "name"]]], "Alice"]
        s = stream({"type": "memory", "data": malformed_data})
        result = s.filter(query)
        assert isinstance(result, FilteredStream)
        assert result.query == query
        # Test evaluation
        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 1  # Only Alice matches
        assert matching_objects[0]["name"] == "Alice"

    def test_input_data_not_a_list_raises_error(self):
        """Test that stream handles non-list data appropriately."""
        # With streams, we can create a source but it might fail during evaluation
        # The behavior depends on how the memory source handles non-list data
        s = stream({"type": "memory", "data": {"not": "a list"}})
        result = s.filter(["eq?", ["@", [["key", "name"]]], "Alice"])
        # Evaluation might fail or return empty results
        # The exact behavior depends on the streaming implementation



"""
Codata/infinite collection loaders for JAF.

These loaders generate potentially infinite streams of data, perfect for
testing lazy evaluation and demonstrating the power of streaming processing.
"""

import random
import itertools
import time
import math
from typing import Generator, Dict, Any, Optional, List
from datetime import datetime, timedelta


def stream_prng(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Stream pseudo-random generated data based on a seed.

    Args:
        source: Dict with:
            - seed: Random seed (optional, defaults to None)
            - template: Template for generated objects
            - limit: Optional limit on number of items (for testing)

    Example:
        {
            "type": "prng",
            "seed": 42,
            "template": {
                "type": "user",
                "id": {"$random": "int", "min": 1, "max": 1000000},
                "name": {"$random": "choice", "choices": ["Alice", "Bob", "Charlie", "Diana"]},
                "age": {"$random": "int", "min": 18, "max": 80},
                "score": {"$random": "float", "min": 0, "max": 100},
                "active": {"$random": "bool"},
                "tags": {"$random": "sample", "population": ["admin", "user", "premium", "beta"], "k": 2}
            }
        }
    """
    seed = source.get("seed")
    template = source.get("template", {})
    limit = source.get("limit")

    rng = random.Random(seed)
    count = 0

    while limit is None or count < limit:
        obj = _generate_from_template(template, rng)
        yield obj
        count += 1


def _generate_from_template(template: Any, rng: random.Random) -> Any:
    """Recursively generate data from a template."""
    if isinstance(template, dict):
        if "$random" in template:
            # Handle random generation directive
            rand_type = template["$random"]

            if rand_type == "int":
                return rng.randint(template.get("min", 0), template.get("max", 100))
            elif rand_type == "float":
                return rng.uniform(template.get("min", 0), template.get("max", 1))
            elif rand_type == "bool":
                return rng.choice([True, False])
            elif rand_type == "choice":
                return rng.choice(template["choices"])
            elif rand_type == "sample":
                population = template["population"]
                k = min(template.get("k", 1), len(population))
                return rng.sample(population, k)
            elif rand_type == "string":
                length = template.get("length", 10)
                chars = template.get("chars", "abcdefghijklmnopqrstuvwxyz")
                return "".join(rng.choice(chars) for _ in range(length))
            elif rand_type == "uuid":
                # Simple UUID-like string
                return f"{rng.randint(0, 0xffffffff):08x}-{rng.randint(0, 0xffff):04x}-{rng.randint(0, 0xffff):04x}-{rng.randint(0, 0xffff):04x}-{rng.randint(0, 0xffffffffffff):012x}"
        else:
            # Regular dict - recursively process values
            return {k: _generate_from_template(v, rng) for k, v in template.items()}
    elif isinstance(template, list):
        # Process each item in the list
        return [_generate_from_template(item, rng) for item in template]
    else:
        # Literal value
        return template


def stream_fibonacci(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Stream Fibonacci sequence as objects.

    Args:
        source: Dict with:
            - start_index: Starting index in sequence (default: 0)
            - include_metadata: Whether to include additional metadata (default: True)
    """
    start_index = source.get("start_index", 0)
    include_metadata = source.get("include_metadata", True)

    a, b = 0, 1
    index = 0

    # Skip to start_index if needed
    for _ in range(start_index):
        a, b = b, a + b
        index += 1

    while True:
        if include_metadata:
            yield {
                "index": index,
                "value": a,
                "ratio_to_previous": b / a if a != 0 else None,
                "is_even": a % 2 == 0,
                "digits": len(str(a)),
                "sum_of_digits": sum(int(d) for d in str(a)),
            }
        else:
            yield {"index": index, "value": a}

        a, b = b, a + b
        index += 1


def stream_time_series(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Stream time series data with various patterns.

    Args:
        source: Dict with:
            - start_time: ISO timestamp or "now" (default: "now")
            - interval_seconds: Seconds between data points (default: 60)
            - pattern: "sine", "random_walk", "sawtooth", "square" (default: "sine")
            - noise: Amount of random noise to add (default: 0.1)
            - trend: Linear trend to add per step (default: 0)
    """
    pattern = source.get("pattern", "sine")
    interval = source.get("interval_seconds", 60)
    noise_level = source.get("noise", 0.1)
    trend = source.get("trend", 0)

    start_time = source.get("start_time", "now")
    if start_time == "now":
        current_time = datetime.now()
    else:
        current_time = datetime.fromisoformat(start_time)

    rng = random.Random(source.get("seed"))
    step = 0
    base_value = 50
    current_value = base_value

    while True:
        # Calculate pattern value
        if pattern == "sine":
            pattern_value = base_value + 20 * math.sin(step * 0.1)
        elif pattern == "sawtooth":
            pattern_value = base_value + 20 * ((step % 20) / 10 - 1)
        elif pattern == "square":
            pattern_value = base_value + 20 * (1 if (step // 10) % 2 == 0 else -1)
        elif pattern == "random_walk":
            current_value += rng.uniform(-2, 2)
            pattern_value = current_value
        else:
            pattern_value = base_value

        # Add trend and noise
        value = (
            pattern_value
            + trend * step
            + rng.uniform(-noise_level * 10, noise_level * 10)
        )

        yield {
            "timestamp": current_time.isoformat(),
            "value": value,
            "step": step,
            "day_of_week": current_time.strftime("%A"),
            "hour": current_time.hour,
            "is_weekend": current_time.weekday() >= 5,
        }

        current_time += timedelta(seconds=interval)
        step += 1


def stream_fractal(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Stream points from fractal patterns (Mandelbrot set, Julia set, etc.).

    Args:
        source: Dict with:
            - fractal_type: "mandelbrot", "julia" (default: "mandelbrot")
            - resolution: Points per unit (default: 100)
            - bounds: {"xmin": -2, "xmax": 2, "ymin": -2, "ymax": 2}
            - max_iterations: Maximum iterations for escape (default: 100)
            - julia_c: Complex constant for Julia set (default: -0.7+0.27i)
    """
    fractal_type = source.get("fractal_type", "mandelbrot")
    resolution = source.get("resolution", 100)
    bounds = source.get("bounds", {"xmin": -2, "xmax": 2, "ymin": -2, "ymax": 2})
    max_iter = source.get("max_iterations", 100)

    # For Julia sets
    julia_c = source.get("julia_c", {"real": -0.7, "imag": 0.27015})
    c_julia = complex(julia_c["real"], julia_c["imag"])

    xmin, xmax = bounds["xmin"], bounds["xmax"]
    ymin, ymax = bounds["ymin"], bounds["ymax"]

    # Generate points in a spiral pattern for better visualization
    for i in itertools.count():
        # Convert linear index to 2D coordinates in a expanding pattern
        ring = int(math.sqrt(i))
        pos_in_ring = i - ring * ring

        if ring == 0:
            x_idx, y_idx = 0, 0
        else:
            side_length = 2 * ring
            if pos_in_ring < side_length:
                x_idx = ring
                y_idx = pos_in_ring - ring
            elif pos_in_ring < 2 * side_length:
                x_idx = ring - (pos_in_ring - side_length)
                y_idx = ring
            elif pos_in_ring < 3 * side_length:
                x_idx = -ring
                y_idx = ring - (pos_in_ring - 2 * side_length)
            else:
                x_idx = -ring + (pos_in_ring - 3 * side_length)
                y_idx = -ring

        # Convert to fractal coordinates
        x = xmin + (x_idx + resolution) * (xmax - xmin) / (2 * resolution)
        y = ymin + (y_idx + resolution) * (ymax - ymin) / (2 * resolution)

        # Skip if outside bounds
        if x < xmin or x > xmax or y < ymin or y > ymax:
            continue

        # Calculate fractal value
        if fractal_type == "mandelbrot":
            c = complex(x, y)
            z = 0
        else:  # julia
            z = complex(x, y)
            c = c_julia

        iterations = 0
        while abs(z) < 2 and iterations < max_iter:
            z = z * z + c
            iterations += 1

        yield {
            "x": x,
            "y": y,
            "iterations": iterations,
            "escaped": iterations < max_iter,
            "convergence_rate": iterations / max_iter,
            "magnitude": abs(z),
        }


def stream_composite(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Stream data by composing/transforming data from other sources.

    Args:
        source: Dict with:
            - sources: List of source descriptors to combine
            - mode: "zip", "round_robin", "merge" (default: "zip")
            - transform: Optional template to transform combined data
    """
    sources = source.get("sources", [])
    mode = source.get("mode", "zip")
    transform = source.get("transform")

    if not sources:
        return

    # Create streams for each source
    streams = [loader.stream(src) for src in sources]

    if mode == "zip":
        # Combine items from each stream together
        for items in itertools.zip_longest(*streams, fillvalue=None):
            if transform:
                yield _apply_transform(items, transform)
            else:
                yield {"sources": [item for item in items if item is not None]}

    elif mode == "round_robin":
        # Alternate between sources
        for item in itertools.chain.from_iterable(
            itertools.zip_longest(*streams, fillvalue=None)
        ):
            if item is not None:
                yield item

    elif mode == "merge":
        # Merge all items into single stream (may exhaust memory if sources are infinite!)
        for stream in streams:
            for item in stream:
                yield item


def _apply_transform(items: List[Any], transform: Dict[str, Any]) -> Dict[str, Any]:
    """Apply a transformation template to combined items."""
    result = {}
    for key, spec in transform.items():
        if isinstance(spec, dict) and "$source" in spec:
            source_idx = spec["$source"]
            field = spec.get("field")
            if 0 <= source_idx < len(items) and items[source_idx] is not None:
                if field:
                    result[key] = items[source_idx].get(field)
                else:
                    result[key] = items[source_idx]
        else:
            result[key] = spec
    return result


def stream_prime(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Stream prime numbers with various properties.

    Args:
        source: Dict with:
            - start: Starting number to check (default: 2)
            - include_gaps: Include gap to previous prime (default: True)
            - include_factors: Include prime factorization (default: False)
    """
    start = source.get("start", 2)
    include_gaps = source.get("include_gaps", True)
    include_factors = source.get("include_factors", False)

    def is_prime(n):
        if n < 2:
            return False
        for i in range(2, int(n**0.5) + 1):
            if n % i == 0:
                return False
        return True

    current = max(2, start)
    prev_prime = None
    prime_count = 0

    while True:
        if is_prime(current):
            prime_count += 1

            result = {
                "value": current,
                "index": prime_count,
                "is_twin_prime": is_prime(current + 2)
                or (current > 2 and is_prime(current - 2)),
                "digit_sum": sum(int(d) for d in str(current)),
                "last_digit": current % 10,
            }

            if include_gaps and prev_prime is not None:
                result["gap"] = current - prev_prime
                result["gap_is_even"] = (current - prev_prime) % 2 == 0

            if include_factors:
                # For primes, the only factors are 1 and itself
                result["factors"] = [1, current]
                result["factor_count"] = 2

            yield result
            prev_prime = current

        current += 1


def stream_distribution(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Stream numbers from various probability distributions.

    Args:
        source: Dict with:
            - distribution: Name of distribution (uniform, normal, exponential, poisson, etc.)
            - seed: Random seed for reproducibility (optional)
            - limit: Optional limit on number of values
            - as_object: If True, yield {"value": x, "index": i}, else just x (default: True)
            - parameters: Distribution-specific parameters

    Supported distributions:
        - uniform: min, max (default: 0, 1)
        - normal/gaussian: mean, std (default: 0, 1)
        - exponential: lambda/rate (default: 1.0)
        - poisson: lambda/mean (default: 1.0)
        - binomial: n, p (default: 10, 0.5)
        - gamma: alpha/shape, beta/scale (default: 1.0, 1.0)
        - beta: alpha, beta (default: 1.0, 1.0)
        - lognormal: mean, std (of underlying normal) (default: 0, 1)
        - chi2: df (degrees of freedom) (default: 1)
        - pareto: alpha/shape (default: 1.0)
        - weibull: shape, scale (default: 1.0, 1.0)

    Example:
        {
            "type": "distribution",
            "distribution": "normal",
            "seed": 42,
            "parameters": {"mean": 100, "std": 15},
            "as_object": true
        }
    """
    dist_name = source.get("distribution", "uniform").lower()
    seed = source.get("seed")
    limit = source.get("limit")
    as_object = source.get("as_object", True)
    params = source.get("parameters", {})

    rng = random.Random(seed)
    count = 0

    # Distribution generators
    if dist_name in ["uniform", "unif"]:
        min_val = params.get("min", 0)
        max_val = params.get("max", 1)
        generator = lambda: rng.uniform(min_val, max_val)

    elif dist_name in ["normal", "gaussian", "norm"]:
        mean = params.get("mean", params.get("mu", 0))
        std = params.get("std", params.get("sigma", 1))
        generator = lambda: rng.gauss(mean, std)

    elif dist_name in ["exponential", "exp"]:
        rate = params.get("lambda", params.get("rate", 1.0))
        generator = lambda: rng.expovariate(rate)

    elif dist_name == "poisson":
        mean = params.get("lambda", params.get("mean", 1.0))

        # Simple Poisson using Knuth's algorithm for small lambda
        def poisson():
            L = math.exp(-mean)
            k = 0
            p = 1.0
            while p > L:
                k += 1
                p *= rng.random()
            return k - 1

        generator = poisson

    elif dist_name == "binomial":
        n = params.get("n", 10)
        p = params.get("p", 0.5)
        generator = lambda: sum(1 for _ in range(n) if rng.random() < p)

    elif dist_name == "gamma":
        alpha = params.get("alpha", params.get("shape", 1.0))
        beta = params.get("beta", params.get("scale", 1.0))
        generator = lambda: rng.gammavariate(alpha, beta)

    elif dist_name == "beta":
        alpha = params.get("alpha", 1.0)
        beta = params.get("beta", 1.0)
        generator = lambda: rng.betavariate(alpha, beta)

    elif dist_name == "lognormal":
        mean = params.get("mean", params.get("mu", 0))
        std = params.get("std", params.get("sigma", 1))
        generator = lambda: rng.lognormvariate(mean, std)

    elif dist_name in ["chi2", "chisquare"]:
        df = params.get("df", 1)
        # Chi-square is gamma with alpha=df/2, beta=2
        generator = lambda: rng.gammavariate(df / 2, 2)

    elif dist_name == "pareto":
        alpha = params.get("alpha", params.get("shape", 1.0))
        generator = lambda: rng.paretovariate(alpha)

    elif dist_name == "weibull":
        shape = params.get("shape", params.get("k", 1.0))
        scale = params.get("scale", params.get("lambda", 1.0))
        generator = lambda: scale * rng.weibullvariate(scale, shape)

    else:
        raise ValueError(f"Unknown distribution: {dist_name}")

    # Generate values
    while limit is None or count < limit:
        value = generator()

        if as_object:
            yield {
                "index": count,
                "value": value,
                "distribution": dist_name,
            }
        else:
            yield value

        count += 1


def stream_counter(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Stream a simple counter/sequence.

    Args:
        source: Dict with:
            - start: Starting value (default: 0)
            - step: Increment step (default: 1)
            - limit: Optional limit
            - as_object: If True, yield {"index": i, "value": v}, else just v
    """
    start = source.get("start", 0)
    step = source.get("step", 1)
    limit = source.get("limit")
    as_object = source.get("as_object", False)

    current = start
    index = 0

    while limit is None or index < limit:
        if as_object:
            yield {"index": index, "value": current}
        else:
            yield current

        current += step
        index += 1


# Register the codata loaders
def register_codata_loaders(loader: "StreamingLoader"):
    """Register all codata loaders with the streaming loader."""
    loader.register("prng", stream_prng)
    loader.register("distribution", stream_distribution)
    loader.register("counter", stream_counter)
    loader.register("fibonacci", stream_fibonacci)
    loader.register("time_series", stream_time_series)
    loader.register("fractal", stream_fractal)
    loader.register("composite", stream_composite)
    loader.register("prime", stream_prime)



"""
Provides a flexible data loading system for JAF collections.

This module introduces the "CollectionLoader," a class that implements a
strategy pattern for loading collections of JSON documents. The core motivation
is to decouple the lazy streams, which are pure data containers representing
filter results, from the implementation details of how and where the original
data is stored.

By using a loader, JAF can be easily extended to support new data sources
(e.g., databases, web APIs, different file formats) without modifying the core
filtering or stream logic. A stream only needs to store a serializable
`collection_source` dictionary, and the `CollectionLoader` uses this metadata
to dispatch to the correct loading function at runtime.

Key Components:
- JsonValue: A type alias representing any valid JSON type.
- LoaderFunc: A type alias for a function that can load a collection.
- CollectionLoader: A class that registers and dispatches to LoaderFuncs.
"""

import json
import gzip
from pathlib import Path
from typing import Any, Callable, Dict, List, Union, Generator

from jaf.path_evaluation import eval_path

# A type alias for any value that can be represented in JSON. This correctly
# reflects the data model JAF operates on, where a "document" can be any
# valid JSON value (object, array, string, number, etc.).
JsonValue = Union[Dict[str, Any], List[Any], str, int, float, bool, None]

# A loader function is a callable that accepts a loader instance and a
# `collection_source` metadata dictionary and returns data.
LoaderFunc = Callable[["CollectionLoader", Dict[str, Any]], Any]

# A streaming loader function yields items one by one
StreamLoaderFunc = Callable[
    ["CollectionLoader", Dict[str, Any]], Generator[JsonValue, None, None]
]


class CollectionLoader:
    """
    Manages and dispatches data loading for different collection source types.

    This class acts as a registry for different data loading strategies. It maps
    a `source_type` string (e.g., "jsonl", "directory") to a specific function
    that knows how to handle that source. The streaming system will delegate
    the task of data retrieval to an instance of this class.
    """

    def __init__(self):
        """Initializes the CollectionLoader and registers default loaders."""
        self._loaders: Dict[str, LoaderFunc] = {}
        self._register_default_loaders()

    def register_loader(self, source_type: str, func: LoaderFunc):
        """Registers or overwrites a loader function for a given source type."""
        self._loaders[source_type] = func

    def load(self, collection_source: Dict[str, Any]) -> List[JsonValue]:
        """
        Loads a collection of documents based on its source metadata.

        This is the main dispatch method. It inspects the 'type' key in the
        `collection_source` dictionary and calls the corresponding registered
        loader function.

        Args:
            collection_source: The source metadata dictionary from a stream.

        Returns:
            A list of documents from the specified source.

        Raises:
            ValueError: If the source type is missing, unknown, or if the
                        source metadata is invalid for the loader.
            IOError: If a file source cannot be read.
        """
        source_type = collection_source.get("source_type") or collection_source.get(
            "type"
        )
        if not source_type:
            raise ValueError(
                "Collection source is missing 'source_type' or 'type' key."
            )

        loader_fn = self._loaders.get(source_type)
        if not loader_fn:
            raise ValueError(f"No loader registered for source type: '{source_type}'")

        return loader_fn(self, collection_source)

    def _register_default_loaders(self):
        """Pre-populates the loader with standard JAF source types."""
        # Compositional loaders
        self.register_loader("path_source", _load_from_path_source)
        self.register_loader("gzip_parser", _parse_gzip)
        self.register_loader("jsonl_parser", _parse_jsonl)
        self.register_loader("json_array_parser", _parse_json_array)

        # Legacy loaders
        self.register_loader("directory", _load_from_directory)
        self.register_loader("jsonl", _legacy_load_from_path)
        self.register_loader("json_array", _legacy_load_from_path)
        self.register_loader("buffered_stdin", _load_from_buffered_stdin)


# --- Compositional Loaders ---


def _load_from_path_source(loader: "CollectionLoader", source: dict) -> bytes:
    paths = source.get("paths", [])
    if len(paths) != 1:
        raise ValueError(
            "path_source with compositional parsing currently supports only one path."
        )

    path = Path(paths[0])
    if not path.is_file():
        raise FileNotFoundError(f"Data source file not found: {path}")
    return path.read_bytes()


def _parse_gzip(loader: "CollectionLoader", source: dict) -> bytes:
    inner_source = source.get("inner_source")
    if not inner_source:
        raise ValueError("gzip_parser requires an inner_source")

    gzipped_bytes = loader.load(inner_source)
    if not isinstance(gzipped_bytes, bytes):
        raise ValueError(
            f"gzip_parser expects bytes from inner_source, got {type(gzipped_bytes)}"
        )

    return gzip.decompress(gzipped_bytes)


def _parse_jsonl(loader: "CollectionLoader", source: dict) -> List[JsonValue]:
    inner_source = source.get("inner_source")
    if not inner_source:
        raise ValueError("jsonl_parser requires an inner_source")

    data = loader.load(inner_source)
    if isinstance(data, bytes):
        data = data.decode("utf-8")

    content = data.strip()
    if not content:
        return []

    return [json.loads(line) for line in content.splitlines() if line.strip()]


def _parse_json_array(loader: "CollectionLoader", source: dict) -> List[JsonValue]:
    inner_source = source.get("inner_source")
    if not inner_source:
        raise ValueError("json_array_parser requires an inner_source")

    data = loader.load(inner_source)
    if isinstance(data, bytes):
        data = data.decode("utf-8")

    content = data.strip()
    if not content:
        return []

    parsed_data = json.loads(content)
    if isinstance(parsed_data, list):
        return parsed_data
    else:
        return [parsed_data]


# --- Legacy Loader Implementations ---


def _load_from_buffered_stdin(loader: "CollectionLoader", source: dict) -> List[Any]:
    """Loader for data buffered directly in the collection source."""
    return source.get("content", [])


def _legacy_load_from_path(
    loader: "CollectionLoader", source: Dict[str, Any]
) -> List[JsonValue]:
    """
    Helper to load a collection from a single file path.

    It intelligently handles both JSONL (one JSON document per line) and
    standard files containing a single JSON array.
    """
    path_str = source.get("path")
    if not path_str:
        raise ValueError("Path-based source is missing 'path' key.")

    path = Path(path_str)
    if not path.is_file():
        raise FileNotFoundError(f"Data source file not found: {path}")

    if path_str.endswith(".gz"):
        with gzip.open(path, "rt", encoding="utf-8") as f:
            content = f.read().strip()
    else:
        content = path.read_text(encoding="utf-8").strip()

    if not content:
        return []

    # First, try to parse the entire content as a single JSON value.
    try:
        data = json.loads(content)
        if isinstance(data, list):
            # It's a file with a JSON array. This is a valid collection.
            return data
        else:
            # It's a file with a single JSON document that is not an array.
            return [data]
    except json.JSONDecodeError:
        # If parsing the whole content fails, it might be a multi-line JSONL file.
        try:
            return [json.loads(line) for line in content.splitlines() if line.strip()]
        except json.JSONDecodeError as e:
            # If both attempts fail, the format is invalid.
            raise ValueError(
                f"File is not a valid JSONL or a file containing a single JSON array: {path}"
            ) from e


def _load_from_directory(loader: "CollectionLoader", source: dict) -> List[Any]:
    """Loader for 'directory' source type."""
    all_objects: List[JsonValue] = []
    file_paths = source.get("files")
    if not file_paths:
        raise ValueError("Directory source is missing 'files' key.")

    for file_path in file_paths:
        try:
            all_objects.extend(_legacy_load_from_path(loader, {"path": file_path}))
        except (IOError, ValueError, FileNotFoundError) as e:
            # Propagate error with context about which file failed.
            raise IOError(
                f"Failed to load or parse file '{file_path}' in directory source: {e}"
            ) from e
    return all_objects



"""
Test the new lazy streaming API with composable operations.
"""

import pytest
from jaf.lazy_streams import stream, FilteredStream, MappedStream


class TestLazyStreams:
    """Test lazy streaming operations"""

    def test_basic_stream_creation(self):
        """Test creating streams from various sources"""
        # From memory
        s = stream({"type": "memory", "data": [1, 2, 3, 4, 5]})
        items = list(s.evaluate())
        assert items == [1, 2, 3, 4, 5]

        # From fibonacci
        s = stream({"type": "fibonacci"})
        items = list(s.take(5).evaluate())
        assert len(items) == 5
        assert items[0]["value"] == 0
        assert items[1]["value"] == 1

    def test_filter_operation(self):
        """Test filtering streams"""
        data = [
            {"name": "Alice", "age": 30},
            {"name": "Bob", "age": 25},
            {"name": "Charlie", "age": 35},
        ]

        s = stream({"type": "memory", "data": data})
        filtered = s.filter(["gt?", ["@", [["key", "age"]]], 30])

        assert isinstance(filtered, FilteredStream)

        items = list(filtered.evaluate())
        assert len(items) == 1
        assert items[0]["name"] == "Charlie"

    def test_map_operation(self):
        """Test mapping/transforming streams"""
        data = [{"name": "Alice", "age": 30}, {"name": "Bob", "age": 25}]

        s = stream({"type": "memory", "data": data})
        mapped = s.map(["@", [["key", "name"]]])

        assert isinstance(mapped, MappedStream)

        items = list(mapped.evaluate())
        assert items == ["Alice", "Bob"]

    def test_composable_operations(self):
        """Test chaining multiple operations"""
        data = [
            {"name": "Alice", "age": 30, "active": True},
            {"name": "Bob", "age": 25, "active": False},
            {"name": "Charlie", "age": 35, "active": True},
            {"name": "Diana", "age": 28, "active": True},
        ]

        # Filter active users, map to names, take first 2
        s = stream({"type": "memory", "data": data})
        result = (
            s.filter(["eq?", ["@", [["key", "active"]]], True])
            .map(["@", [["key", "name"]]])
            .take(2)
        )

        items = list(result.evaluate())
        assert items == ["Alice", "Charlie"]

    def test_filter_after_map(self):
        """Test filtering after transformation"""
        data = [
            {"user": {"name": "Alice", "score": 95}},
            {"user": {"name": "Bob", "score": 75}},
            {"user": {"name": "Charlie", "score": 85}},
        ]

        s = stream({"type": "memory", "data": data})

        # Map to extract user objects, then filter by score
        result = s.map(["@", [["key", "user"]]]).filter(
            ["gt?", ["@", [["key", "score"]]], 80]
        )

        items = list(result.evaluate())
        assert len(items) == 2
        assert {item["name"] for item in items} == {"Alice", "Charlie"}

    def test_complex_transformation(self):
        """Test complex transformations with dict construction"""
        data = [
            {"name": "Alice", "age": 30, "email": "alice@example.com"},
            {"name": "Bob", "age": 25, "email": "bob@example.com"},
        ]

        s = stream({"type": "memory", "data": data})

        # Transform to new structure
        result = s.map(
            [
                "dict",
                "display_name",
                ["upper-case", ["@", [["key", "name"]]]],
                "contact",
                ["@", [["key", "email"]]],
                "adult",
                ["gte?", ["@", [["key", "age"]]], 18],
            ]
        )

        items = list(result.evaluate())
        assert len(items) == 2
        assert items[0]["display_name"] == "ALICE"
        assert items[0]["contact"] == "alice@example.com"
        assert items[0]["adult"] is True

    def test_stream_filter_convenience(self):
        """Test creating a stream and immediately filtering"""
        data = [
            {"name": "Alice", "age": 30},
            {"name": "Bob", "age": 25},
            {"name": "Charlie", "age": 35},
        ]

        # Create stream and filter
        s = stream({"type": "memory", "data": data})
        filtered = s.filter(["gt?", ["@", [["key", "age"]]], 30])

        # Can chain operations
        result = filtered.map(["@", [["key", "name"]]])

        items = list(result.evaluate())
        assert items == ["Charlie"]

    def test_infinite_stream_operations(self):
        """Test operations on infinite streams"""
        # Fibonacci numbers
        s = stream({"type": "fibonacci"})

        # Take first 10, filter even values, map to just the value
        result = (
            s.take(10)
            .filter(["eq?", ["@", [["key", "is_even"]]], True])
            .map(["@", [["key", "value"]]])
        )

        items = list(result.evaluate())
        # Even fibonacci numbers in first 10: 0, 2, 8, 34
        assert items == [0, 2, 8, 34]

    def test_slice_operation(self):
        """Test slicing streams"""
        s = stream({"type": "fibonacci"})

        # Get fibonacci numbers from index 5 to 10
        result = s.slice(5, 10).map(["@", [["key", "value"]]])

        items = list(result.evaluate())
        assert items == [5, 8, 13, 21, 34]  # F(5) through F(9)

    def test_pipeline_serialization(self):
        """Test that pipelines can be serialized"""
        data = [{"x": 1}, {"x": 2}, {"x": 3}]

        s = stream({"type": "memory", "data": data})
        pipeline = (
            s.filter(["gt?", ["@", [["key", "x"]]], 1])
            .map(["*", ["@", [["key", "x"]]], 10])
            .take(1)
        )

        # The pipeline should be serializable
        pipeline_dict = pipeline.to_dict()
        assert pipeline_dict["collection_source"]["type"] == "take"
        assert pipeline_dict["collection_source"]["n"] == 1

        # The inner source should be a map
        inner = pipeline_dict["collection_source"]["inner_source"]
        assert inner["type"] == "map"

        # And its inner source should be a filter
        inner_inner = inner["inner_source"]
        assert inner_inner["type"] == "filter"



"""
JAF Path Operations
===================

This module contains the path operation dispatcher and all path operation handlers.
"""

import logging
import re
from typing import Any, List
import rapidfuzz.distance as distance
import fuzzy
from .path_exceptions import PathSyntaxError

logger = logging.getLogger(__name__)


def _fuzzy_match_keys(
    target_key: str, available_keys: List[str], cutoff: float, algorithm: str
) -> List[str]:
    """
    Find keys that fuzzy match the target key.

    Args:
        target_key: The key to match against
        available_keys: List of available keys to search
        cutoff: Minimum similarity score (0.0 to 1.0)
        algorithm: Algorithm to use ('difflib', 'levenshtein', 'jaro_winkler', 'soundex', 'metaphone')

    Returns:
        List of matching keys
    """
    if algorithm == "difflib":
        import difflib

        matches = difflib.get_close_matches(target_key, available_keys, cutoff=cutoff)
        return matches
    elif algorithm == "levenshtein":
        matches = []
        for key in available_keys:
            similarity = 1 - (
                distance.Levenshtein.distance(target_key, key)
                / max(len(target_key), len(key))
            )
            if similarity >= cutoff:
                matches.append(key)
        return matches
    elif algorithm == "jaro_winkler":
        matches = []
        for key in available_keys:
            similarity = distance.JaroWinkler.similarity(target_key, key)
            if similarity >= cutoff:
                matches.append(key)
        return matches
    elif algorithm == "metaphone":
        target_metaphone = fuzzy.DMetaphone()(target_key)[0]
        matches = []
        for key in available_keys:
            key_metaphone = fuzzy.DMetaphone()(key)[0]
            if target_metaphone and target_metaphone == key_metaphone:
                matches.append(key)
        return matches
    else:
        raise ValueError(f"Unknown fuzzy matching algorithm: {algorithm}")


class PathOperationDispatcher:
    """Dispatcher for path operations"""

    def __init__(self):
        self.operations = {
            "key": self._handle_key,
            "index": self._handle_index,
            "indices": self._handle_indices,
            "slice": self._handle_slice,
            "regex_key": self._handle_regex_key,
            "fuzzy_key": self._handle_fuzzy_key,
            "wc_level": self._handle_wc_level,
            "wc_recursive": self._handle_wc_recursive,
            "root": self._handle_root,
        }

    def dispatch(
        self,
        op: str,
        args: List[Any],
        current_obj: Any,
        remaining_components: List[List[Any]],
        full_path_ast_for_error: List[List[Any]],
        root_obj_for_path: Any,
    ) -> List[Any]:
        """Dispatch to appropriate handler"""
        if op not in self.operations:
            raise PathSyntaxError(
                f"Unknown path operation: '{op}'",
                path_segment=[op] + args,
                full_path_ast=full_path_ast_for_error,
            )

        return self.operations[op](
            args,
            current_obj,
            remaining_components,
            full_path_ast_for_error,
            root_obj_for_path,
        )

    def _handle_key(
        self,
        args: List[Any],
        current_obj: Any,
        remaining_components: List[List[Any]],
        full_path_ast_for_error: List[List[Any]],
        root_obj_for_path: Any,
    ) -> List[Any]:
        # Import here to avoid circular imports
        from .path_evaluation import _match_recursive

        if not (len(args) == 1 and isinstance(args[0], str)):
            raise PathSyntaxError(
                "'key' operation expects a single string argument.",
                path_segment=["key"] + args,
                full_path_ast=full_path_ast_for_error,
            )
        key_name = args[0]
        if isinstance(current_obj, dict) and key_name in current_obj:
            return _match_recursive(
                current_obj[key_name],
                remaining_components,
                full_path_ast_for_error,
                root_obj_for_path,
            )
        return []

    def _handle_index(
        self,
        args: List[Any],
        current_obj: Any,
        remaining_components: List[List[Any]],
        full_path_ast_for_error: List[List[Any]],
        root_obj_for_path: Any,
    ) -> List[Any]:
        # Import here to avoid circular imports
        from .path_evaluation import _match_recursive

        if not (len(args) == 1 and isinstance(args[0], int)):
            raise PathSyntaxError(
                "'index' operation expects a single integer argument.",
                path_segment=["index"] + args,
                full_path_ast=full_path_ast_for_error,
            )
        idx_val = args[0]
        if isinstance(current_obj, list):
            if -len(current_obj) <= idx_val < len(current_obj):
                return _match_recursive(
                    current_obj[idx_val],
                    remaining_components,
                    full_path_ast_for_error,
                    root_obj_for_path,
                )
        return []

    def _handle_indices(
        self,
        args: List[Any],
        current_obj: Any,
        remaining_components: List[List[Any]],
        full_path_ast_for_error: List[List[Any]],
        root_obj_for_path: Any,
    ) -> List[Any]:
        # Import here to avoid circular imports
        from .path_evaluation import _match_recursive

        if not (
            len(args) == 1
            and isinstance(args[0], list)
            and all(isinstance(idx, int) for idx in args[0])
        ):
            raise PathSyntaxError(
                "'indices' operation expects a single list of integers argument.",
                path_segment=["indices"] + args,
                full_path_ast=full_path_ast_for_error,
            )
        idx_list = args[0]
        if isinstance(current_obj, list):
            collected_values = []
            for idx_val in idx_list:
                if isinstance(idx_val, int) and -len(current_obj) <= idx_val < len(
                    current_obj
                ):
                    collected_values.extend(
                        _match_recursive(
                            current_obj[idx_val],
                            remaining_components,
                            full_path_ast_for_error,
                            root_obj_for_path,
                        )
                    )
            return collected_values
        return []

    def _handle_slice(
        self,
        args: List[Any],
        current_obj: Any,
        remaining_components: List[List[Any]],
        full_path_ast_for_error: List[List[Any]],
        root_obj_for_path: Any,
    ) -> List[Any]:
        # Import here to avoid circular imports
        from .path_evaluation import _match_recursive

        if not (1 <= len(args) <= 3):
            raise PathSyntaxError(
                "'slice' operation expects 1 to 3 arguments for start, stop, step.",
                path_segment=["slice"] + args,
                full_path_ast=full_path_ast_for_error,
            )

        start_val = args[0]
        stop_val = args[1] if len(args) > 1 else None
        step_val = args[2] if len(args) > 2 else None

        if not (start_val is None or isinstance(start_val, int)):
            raise PathSyntaxError(
                "Slice start must be an integer or null.",
                path_segment=args,
                full_path_ast=full_path_ast_for_error,
            )
        if not (stop_val is None or isinstance(stop_val, int)):
            raise PathSyntaxError(
                "Slice stop must be an integer or null.",
                path_segment=args,
                full_path_ast=full_path_ast_for_error,
            )

        actual_step = step_val if step_val is not None else 1
        if not (isinstance(actual_step, int) and actual_step != 0):
            raise PathSyntaxError(
                "Slice step must be a non-zero integer.",
                path_segment=args,
                full_path_ast=full_path_ast_for_error,
            )

        if isinstance(current_obj, list):
            try:
                s = slice(start_val, stop_val, actual_step)
                sliced_items = current_obj[s]
                collected_values = []
                for item in sliced_items:
                    collected_values.extend(
                        _match_recursive(
                            item,
                            remaining_components,
                            full_path_ast_for_error,
                            root_obj_for_path,
                        )
                    )
                return collected_values
            except (
                TypeError,
                ValueError,
            ) as e:  # Should be rare if AST validation is correct
                logger.debug(
                    f"Error during slicing for {current_obj} with slice({start_val},{stop_val},{actual_step}): {e}"
                )
        return []

    def _handle_regex_key(
        self,
        args: List[Any],
        current_obj: Any,
        remaining_components: List[List[Any]],
        full_path_ast_for_error: List[List[Any]],
        root_obj_for_path: Any,
    ) -> List[Any]:
        # Import here to avoid circular imports
        from .path_evaluation import _match_recursive

        if not (1 <= len(args) <= 2):
            raise PathSyntaxError(
                "'regex_key' operation expects 1 or 2 arguments: pattern, [flags].",
                path_segment=["regex_key"] + args,
                full_path_ast=full_path_ast_for_error,
            )

        if not isinstance(args[0], str):
            raise PathSyntaxError(
                "'regex_key' operation expects a string argument for the pattern.",
                path_segment=["regex_key"] + args,
                full_path_ast=full_path_ast_for_error,
            )

        pattern = args[0]
        flags = 0  # Default no flags

        # Parse optional flags argument
        if len(args) == 2:
            if isinstance(args[1], str):
                # String flags like "i", "m", "s", etc.
                flag_str = args[1].lower()
                for flag_char in flag_str:
                    if flag_char == "i":
                        flags |= re.IGNORECASE
                    elif flag_char == "m":
                        flags |= re.MULTILINE
                    elif flag_char == "s":
                        flags |= re.DOTALL
                    elif flag_char == "x":
                        flags |= re.VERBOSE
                    elif flag_char == "a":
                        flags |= re.ASCII
                    else:
                        raise PathSyntaxError(
                            f"'regex_key' operation: unknown flag '{flag_char}'. Valid flags: i, m, s, x, a.",
                            path_segment=["regex_key"] + args,
                            full_path_ast=full_path_ast_for_error,
                        )
            elif isinstance(args[1], int):
                # Integer flags (direct re module constants)
                flags = args[1]
            else:
                raise PathSyntaxError(
                    "'regex_key' operation expects a string or integer argument for flags.",
                    path_segment=["regex_key"] + args,
                    full_path_ast=full_path_ast_for_error,
                )

        if isinstance(current_obj, dict):
            collected_values = []
            try:
                compiled_pattern = re.compile(pattern, flags)
                for key in current_obj:
                    if compiled_pattern.search(key):
                        collected_values.extend(
                            _match_recursive(
                                current_obj[key],
                                remaining_components,
                                full_path_ast_for_error,
                                root_obj_for_path,
                            )
                        )
            except re.error as e:
                raise PathSyntaxError(
                    f"'regex_key' operation: invalid regex pattern '{pattern}': {e}",
                    path_segment=["regex_key"] + args,
                    full_path_ast=full_path_ast_for_error,
                )
            return collected_values

        return []

    def _handle_wc_level(
        self,
        args: List[Any],
        current_obj: Any,
        remaining_components: List[List[Any]],
        full_path_ast_for_error: List[List[Any]],
        root_obj_for_path: Any,
    ) -> List[Any]:
        # Import here to avoid circular imports
        from .path_evaluation import _match_recursive

        if args:
            raise PathSyntaxError(
                "'wc_level' operation expects no arguments.",
                path_segment=["wc_level"] + args,
                full_path_ast=full_path_ast_for_error,
            )
        if isinstance(current_obj, dict):
            collected_values = []
            for v_obj in current_obj.values():
                collected_values.extend(
                    _match_recursive(
                        v_obj,
                        remaining_components,
                        full_path_ast_for_error,
                        root_obj_for_path,
                    )
                )
            return collected_values
        elif isinstance(current_obj, list):
            collected_values = []
            for item in current_obj:
                collected_values.extend(
                    _match_recursive(
                        item,
                        remaining_components,
                        full_path_ast_for_error,
                        root_obj_for_path,
                    )
                )
            return collected_values
        return []

    def _handle_wc_recursive(
        self,
        args: List[Any],
        current_obj: Any,
        remaining_components: List[List[Any]],
        full_path_ast_for_error: List[List[Any]],
        root_obj_for_path: Any,
    ) -> List[Any]:
        # Import here to avoid circular imports
        from .path_evaluation import _match_recursive

        if args:
            raise PathSyntaxError(
                "'wc_recursive' operation expects no arguments.",
                path_segment=["wc_recursive"] + args,
                full_path_ast=full_path_ast_for_error,
            )

        collected_values = []

        # Match current level against remainder
        collected_values.extend(
            _match_recursive(
                current_obj,
                remaining_components,
                full_path_ast_for_error,
                root_obj_for_path,
            )
        )

        # Match children against the wc_recursive op + remainder
        if remaining_components:  # Only recurse if there are remaining components
            current_recursive_op_and_remainder = [
                ["wc_recursive"]
            ] + remaining_components
            if isinstance(current_obj, dict):
                for v_obj in current_obj.values():
                    collected_values.extend(
                        _match_recursive(
                            v_obj,
                            current_recursive_op_and_remainder,
                            full_path_ast_for_error,
                            root_obj_for_path,
                        )
                    )
            elif isinstance(current_obj, list):
                for item in current_obj:
                    collected_values.extend(
                        _match_recursive(
                            item,
                            current_recursive_op_and_remainder,
                            full_path_ast_for_error,
                            root_obj_for_path,
                        )
                    )

        return collected_values

    def _handle_root(
        self,
        args: List[Any],
        current_obj: Any,
        remaining_components: List[List[Any]],
        full_path_ast_for_error: List[List[Any]],
        root_obj_for_path: Any,
    ) -> List[Any]:
        # Import here to avoid circular imports
        from .path_evaluation import _match_recursive

        if args:
            raise PathSyntaxError(
                "'root' operation expects no arguments.",
                path_segment=["root"] + args,
                full_path_ast=full_path_ast_for_error,
            )
        # Reset current_obj to the absolute root and continue with remaining components
        return _match_recursive(
            root_obj_for_path,
            remaining_components,
            full_path_ast_for_error,
            root_obj_for_path,
        )

    def _handle_fuzzy_key(
        self,
        args: List[Any],
        current_obj: Any,
        remaining_components: List[List[Any]],
        full_path_ast_for_error: List[List[Any]],
        root_obj_for_path: Any,
    ) -> List[Any]:
        # Import here to avoid circular imports
        from .path_evaluation import _match_recursive

        if len(args) not in [1, 2, 3]:
            raise PathSyntaxError(
                "'fuzzy_key' operation expects 1 to 3 arguments: key_name, [cutoff], [algorithm].",
                path_segment=["fuzzy_key"] + args,
                full_path_ast=full_path_ast_for_error,
            )

        if not isinstance(args[0], str):
            raise PathSyntaxError(
                "'fuzzy_key' operation expects a string argument for the key name.",
                path_segment=["fuzzy_key"] + args,
                full_path_ast=full_path_ast_for_error,
            )

        key_name = args[0]
        cutoff = 0.6  # Default fuzzy match cutoff
        algorithm = "difflib"  # Default algorithm

        # Parse optional cutoff argument
        if len(args) >= 2:
            if not isinstance(args[1], (float, int)):
                raise PathSyntaxError(
                    "'fuzzy_key' operation expects a numeric argument for the cutoff.",
                    path_segment=["fuzzy_key"] + args,
                    full_path_ast=full_path_ast_for_error,
                )
            cutoff = float(args[1])
            if not (0.0 <= cutoff <= 1.0):
                raise PathSyntaxError(
                    "'fuzzy_key' operation expects a cutoff between 0.0 and 1.0.",
                    path_segment=["fuzzy_key"] + args,
                    full_path_ast=full_path_ast_for_error,
                )

        # Parse optional algorithm argument
        if len(args) == 3:
            if not isinstance(args[2], str):
                raise PathSyntaxError(
                    "'fuzzy_key' operation expects a string argument for the algorithm.",
                    path_segment=["fuzzy_key"] + args,
                    full_path_ast=full_path_ast_for_error,
                )
            algorithm = args[2].lower()
            valid_algorithms = [
                "difflib",
                "levenshtein",
                "jaro_winkler",
                "soundex",
                "metaphone",
            ]
            if algorithm not in valid_algorithms:
                raise PathSyntaxError(
                    f"'fuzzy_key' operation: unknown algorithm '{algorithm}'. Valid options: {', '.join(valid_algorithms)}.",
                    path_segment=["fuzzy_key"] + args,
                    full_path_ast=full_path_ast_for_error,
                )

        if isinstance(current_obj, dict):
            collected_values = []
            matches = _fuzzy_match_keys(
                key_name, list(current_obj.keys()), cutoff, algorithm
            )
            for matched_key in matches:
                collected_values.extend(
                    _match_recursive(
                        current_obj[matched_key],
                        remaining_components,
                        full_path_ast_for_error,
                        root_obj_for_path,
                    )
                )
            return collected_values

        return []



"""
Tests for lazy evaluation functionality and the evaluate() method.
These tests ensure the lazy evaluation system works correctly and covers edge cases.
"""

import pytest
from jaf.lazy_streams import stream, FilteredStream
from jaf.exceptions import JAFError, UnknownOperatorError, InvalidArgumentCountError


class TestLazyEvaluation:
    """Test lazy evaluation behavior and evaluate() method"""

    def setup_method(self):
        """Set up test data"""
        self.test_data = [
            {
                "id": 1,
                "name": "Alice",
                "age": 30,
                "active": True,
                "tags": ["admin", "user"],
            },
            {"id": 2, "name": "Bob", "age": 25, "active": False, "tags": ["user"]},
            {
                "id": 3,
                "name": "Charlie",
                "age": 35,
                "active": True,
                "tags": ["user", "premium"],
            },
            {"id": 4, "name": "Diana", "age": 28, "active": True, "tags": []},
        ]

    def test_evaluate_basic_functionality(self):
        """Test basic evaluate() functionality"""
        query = ["eq?", ["@", [["key", "active"]]], True]
        s = stream({"type": "memory", "data": self.test_data})
        result = s.filter(query)

        # Query creation should be instant (lazy)
        assert isinstance(result, FilteredStream)
        assert result.query == query

        # Evaluation should return actual matching objects
        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 3  # Alice, Charlie, Diana
        matching_names = {obj["name"] for obj in matching_objects}
        assert matching_names == {"Alice", "Charlie", "Diana"}

    def test_evaluate_is_repeatable(self):
        """Test that evaluate() can be called multiple times with same results"""
        query = ["gt?", ["@", [["key", "age"]]], 30]
        s = stream({"type": "memory", "data": self.test_data})
        result = s.filter(query)

        # First evaluation
        first_eval = list(result.evaluate())
        assert len(first_eval) == 1
        assert first_eval[0]["name"] == "Charlie"

        # Second evaluation should give same results
        second_eval = list(result.evaluate())
        assert first_eval == second_eval

        # Third evaluation should also be consistent
        third_eval = list(result.evaluate())
        assert first_eval == third_eval

    def test_evaluate_with_empty_results(self):
        """Test evaluate() when query matches nothing"""
        query = ["eq?", ["@", [["key", "name"]]], "Nobody"]
        s = stream({"type": "memory", "data": self.test_data})
        result = s.filter(query)

        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 0
        assert matching_objects == []

    def test_evaluate_with_empty_data(self):
        """Test evaluate() with empty data collection"""
        query = ["eq?", ["@", [["key", "name"]]], "Alice"]
        s = stream({"type": "memory", "data": []})
        result = s.filter(query)

        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 0
        assert matching_objects == []

    def test_evaluate_with_complex_query(self):
        """Test evaluate() with complex nested query"""
        query = [
            "and",
            ["eq?", ["@", [["key", "active"]]], True],
            ["gt?", ["@", [["key", "age"]]], 25],
            ["in?", "user", ["@", [["key", "tags"]]]],
        ]
        s = stream({"type": "memory", "data": self.test_data})
        result = s.filter(query)

        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 2  # Alice and Charlie
        matching_names = {obj["name"] for obj in matching_objects}
        assert matching_names == {"Alice", "Charlie"}

    def test_evaluate_with_invalid_query_fails_clearly(self):
        """Test that evaluate() fails clearly for invalid queries"""
        query = ["unknown-operator", "arg"]
        s = stream({"type": "memory", "data": self.test_data})
        result = s.filter(query)

        # Query creation should succeed (lazy)
        assert isinstance(result, FilteredStream)

        # But evaluation should fail with clear error
        with pytest.raises(
            UnknownOperatorError,
            match="Unknown operator: unknown-operator",
        ):
            list(result.evaluate())

    def test_evaluate_with_malformed_data_handles_gracefully(self):
        """Test that evaluate() handles malformed data gracefully"""
        malformed_data = [
            {"id": 1, "name": "Alice"},  # Good
            "not a dict",  # Bad - should be skipped
            {"id": 2, "name": "Bob"},  # Good
            None,  # Bad - should be skipped
            123,  # Bad - should be skipped
            {"id": 3, "name": "Charlie"},  # Good
        ]

        query = ["eq?", ["@", [["key", "name"]]], "Bob"]
        s = stream({"type": "memory", "data": malformed_data})
        result = s.filter(query)

        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 1
        assert matching_objects[0]["name"] == "Bob"

    def test_evaluate_with_missing_fields_handles_gracefully(self):
        """Test that evaluate() handles missing fields gracefully"""
        data_with_missing_fields = [
            {"id": 1, "name": "Alice", "age": 30},  # Has age
            {"id": 2, "name": "Bob"},  # Missing age
            {"id": 3, "name": "Charlie", "age": 35},  # Has age
        ]

        query = ["gt?", ["@", [["key", "age"]]], 25]
        s = stream({"type": "memory", "data": data_with_missing_fields})
        result = s.filter(query)

        # Should only match objects that have the field
        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 2  # Alice and Charlie
        matching_names = {obj["name"] for obj in matching_objects}
        assert matching_names == {"Alice", "Charlie"}

    def test_lazy_boolean_operations_evaluate_correctly(self):
        """Test that boolean operations on queries evaluate correctly"""
        query1 = ["eq?", ["@", [["key", "active"]]], True]
        query2 = ["gt?", ["@", [["key", "age"]]], 30]

        s = stream({"type": "memory", "data": self.test_data})
        rs1 = s.filter(query1)
        rs2 = s.filter(query2)

        # Test AND operation
        and_result = rs1.AND(rs2)
        and_matches = list(and_result.evaluate())
        assert len(and_matches) == 1  # Only Charlie (active AND age > 30)
        assert and_matches[0]["name"] == "Charlie"

        # Test OR operation
        or_result = rs1.OR(rs2)
        or_matches = list(or_result.evaluate())
        assert len(or_matches) == 3  # Alice, Charlie, Diana (active OR age > 30)
        or_names = {obj["name"] for obj in or_matches}
        assert or_names == {"Alice", "Charlie", "Diana"}

        # Test NOT operation
        not_result = rs1.NOT()
        not_matches = list(not_result.evaluate())
        assert len(not_matches) == 1  # Only Bob (not active)
        assert not_matches[0]["name"] == "Bob"

    def test_evaluate_preserves_object_identity(self):
        """Test that evaluate() returns the original objects, not copies"""
        query = ["eq?", ["@", [["key", "name"]]], "Alice"]
        s = stream({"type": "memory", "data": self.test_data})
        result = s.filter(query)

        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 1

        # Should be the same object reference (not a copy)
        original_alice = self.test_data[0]
        evaluated_alice = matching_objects[0]
        assert original_alice is evaluated_alice

    def test_collection_source_metadata_preserved(self):
        """Test that collection source metadata is preserved through evaluation"""
        query = ["eq?", ["@", [["key", "active"]]], True]
        collection_source = {
            "type": "memory",
            "data": self.test_data,
            "description": "test data",
        }

        # For collection_source, we use it directly with stream
        s = stream(collection_source)
        result = s.filter(query)
        result.collection_id = "test_collection"

        # Metadata should be preserved
        assert result.collection_id == "test_collection"
        assert result.collection_source["type"] == "filter"  # Filter wraps the source

        # Evaluation should still work
        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 3

    def test_evaluate_with_collection_source_in_memory(self):
        """Test evaluate() works correctly with in-memory collection source"""
        query = ["eq?", ["@", [["key", "name"]]], "Bob"]

        # This should create in-memory collection source
        s = stream({"type": "memory", "data": self.test_data})
        result = s.filter(query)

        # Should have filter wrapping memory source
        assert result.collection_source["type"] == "filter"
        assert result.collection_source["inner_source"]["type"] == "memory"
        assert result.collection_source["inner_source"]["data"] == self.test_data

        # Evaluation should work
        matching_objects = list(result.evaluate())
        assert len(matching_objects) == 1
        assert matching_objects[0]["name"] == "Bob"


class TestEvaluatePerformance:
    """Test evaluate() performance characteristics"""

    def test_lazy_creation_is_fast(self):
        """Test that query creation is fast (doesn't evaluate immediately)"""
        # Create large dataset
        large_data = [{"id": i, "value": i * 2} for i in range(1000)]

        query = ["gt?", ["@", [["key", "value"]]], 500]

        # Query creation should be instant (no actual filtering)
        s = stream({"type": "memory", "data": large_data})
        result = s.filter(query)
        assert isinstance(result, FilteredStream)
        assert result.query == query

        # Only evaluate() should do the work
        matching_objects = list(result.evaluate())
        assert len(matching_objects) > 0  # Should find matches

    def test_evaluate_multiple_times_performance(self):
        """Test that multiple evaluate() calls don't re-evaluate from scratch"""
        # Note: Current implementation re-evaluates each time
        # This test documents current behavior and can be optimized later
        data = [{"id": i, "active": i % 2 == 0} for i in range(100)]
        query = ["eq?", ["@", [["key", "active"]]], True]

        s = stream({"type": "memory", "data": data})
        result = s.filter(query)

        # Multiple evaluations should work
        eval1 = list(result.evaluate())
        eval2 = list(result.evaluate())
        eval3 = list(result.evaluate())

        # Results should be consistent
        assert eval1 == eval2 == eval3
        assert len(eval1) == 50  # Half should be active



import unittest
from jaf.jaf_eval import jaf_eval
from jaf.path_exceptions import PathSyntaxError
from jaf.exceptions import JAFError


class TestJafEvalPathStrings(unittest.TestCase):
    """Test JAF evaluation with path strings that get converted to AST"""

    def setUp(self):
        """Set up test data"""
        self.test_data = {
            "user": {
                "name": "John Doe",
                "age": 30,
                "email": "john@example.com",
                "tags": ["developer", "python", "javascript"],
                "profile": {
                    "bio": "Software developer",
                    "skills": {
                        "languages": ["Python", "JavaScript", "Go"],
                        "frameworks": ["Django", "React", "Vue"],
                    },
                },
            },
            "user-test-wc": {
                "name": "Jane Doe",
            },
            "items": [
                {"id": 1, "name": "Item 1", "active": True},
                {"id": 2, "name": "Item 2", "active": False},
                {"id": 3, "name": "Item 3", "active": True},
            ],
            "config": {"api_version": "v1.2.3", "debug_mode": True, "timeout": 30},
        }

    def test_simple_key_path_string(self):
        """Test simple key access using path string"""
        # String path
        result = jaf_eval.eval(["@", "user.name"], self.test_data)
        self.assertEqual(result, "John Doe")

    def test_simple_wc(self):
        """Test simple key access using path string"""
        # String path
        # result = jaf_eval.eval(["@", [["wc_level"], ['key', 'name']]], self.test_data)
        result = jaf_eval.eval(["@", "*.name"], self.test_data)

        # check to ensure it has both John Doe and Jane Doe
        self.assertEqual(set(result), {"John Doe", "Jane Doe"})

    def test_array_index_path_string(self):
        """Test array index access using path string"""
        result = jaf_eval.eval(["@", "items[0].name"], self.test_data)
        self.assertEqual(result, "Item 1")

        result = jaf_eval.eval(["@", "items[1].id"], self.test_data)
        self.assertEqual(result, 2)

    def test_array_slice_path_string(self):
        """Test array slicing using path string"""
        result = jaf_eval.eval(["@", "items[0:2].name"], self.test_data)
        self.assertEqual(set(result), {"Item 1", "Item 2"})

        result = jaf_eval.eval(["@", "items[1:].id"], self.test_data)
        self.assertEqual(set(result), {2, 3})

    def test_array_indices_path_string(self):
        """Test specific array indices using path string"""
        result = jaf_eval.eval(["@", "items[0,2].name"], self.test_data)
        self.assertEqual(set(result), {"Item 1", "Item 3"})

    def test_wildcard_level_path_string(self):
        """Test level wildcard using path string"""
        result = jaf_eval.eval(["@", "items[*].name"], self.test_data)
        self.assertEqual(set(result), {"Item 1", "Item 2", "Item 3"})

        # Test explicit bracket form
        result2 = jaf_eval.eval(["@", "items[*].id"], self.test_data)
        self.assertEqual(set(result2), {1, 2, 3})

    def test_wildcard_recursive_path_string(self):
        """Test recursive wildcard using path string"""
        result = jaf_eval.eval(["@", "**.name"], self.test_data)
        expected_names = {"John Doe", "Jane Doe", "Item 1", "Item 2", "Item 3"}
        self.assertEqual(set(result), expected_names)

    def test_complex_nested_path_string(self):
        """Test complex nested path with wildcards"""
        result = jaf_eval.eval(["@", "user.profile.skills[*][*]"], self.test_data)
        expected_skills = {"Python", "JavaScript", "Go", "Django", "React", "Vue"}
        self.assertEqual(set(result), expected_skills)

    def test_path_string_vs_ast_equivalence(self):
        """Test that string paths and equivalent AST paths produce the same results"""
        test_cases = [
            ("user.name", [["key", "user"], ["key", "name"]]),
            ("items[0]", [["key", "items"], ["index", 0]]),
            ("items[*].name", [["key", "items"], ["wc_level"], ["key", "name"]]),
            ("items[0:2]", [["key", "items"], ["slice", 0, 2]]),
            ("items[0,2]", [["key", "items"], ["indices", [0, 2]]]),
        ]

        for path_string, path_ast in test_cases:
            with self.subTest(path_string=path_string):
                string_result = jaf_eval.eval(["@", path_string], self.test_data)
                ast_result = jaf_eval.eval(["@", path_ast], self.test_data)
                self.assertEqual(string_result, ast_result)

    def test_path_string_with_exists_predicate_1(self):
        """Test path strings used within exists? predicate"""
        # Test with existing path
        result = jaf_eval.eval(["exists?", ["@", "user.email"]], self.test_data)
        self.assertTrue(result)

    def test_path_string_with_exists_predicate_2(self):
        # Test with non-existing path
        result = jaf_eval.eval(["exists?", ["@", "user.phone"]], self.test_data)
        print(f"Result for non-existing path: {result}")
        self.assertFalse(result)

    def test_path_string_with_exists_predicate_3(self):
        # Test with array element
        result = jaf_eval.eval(["exists?", ["@", "items[0].name"]], self.test_data)
        self.assertTrue(result)

    def test_path_string_in_filter_expressions(self):
        """Test path strings used in complex filter expressions"""
        # Check if user's age is greater than 25
        result = jaf_eval.eval(["gt?", ["@", "user.age"], 25], self.test_data)
        self.assertTrue(result)

        # Check if first item is active
        result = jaf_eval.eval(["eq?", ["@", "items[0].active"], True], self.test_data)
        self.assertTrue(result)

    def test_path_string_with_conditional_logic(self):
        """Test path strings in if/and/or expressions"""
        # If user age > 25, return user name, otherwise return "Unknown"
        result = jaf_eval.eval(
            ["if", ["gt?", ["@", "user.age"], 25], ["@", "user.name"], "Unknown"],
            self.test_data,
        )
        self.assertEqual(result, "John Doe")

        # Check if user exists and has email
        result = jaf_eval.eval(
            ["and", ["exists?", ["@", "user"]], ["exists?", ["@", "user.email"]]],
            self.test_data,
        )
        self.assertTrue(result)

    def test_invalid_path_string_syntax(self):
        """Test error handling for invalid path string syntax"""
        invalid_paths = [
            "user[",  # Unterminated bracket
            "user..",  # Double dot
            "user[]",  # Empty brackets
            "user[abc]",  # Invalid index
            "user[1:2:3:4]",  # Too many colons
            "~/unterminated",  # Unterminated regex
        ]

        for invalid_path in invalid_paths:
            with self.subTest(path=invalid_path):
                with self.assertRaises(PathSyntaxError):
                    jaf_eval.eval(["@", invalid_path], self.test_data)

    def test_path_argument_validation(self):
        """Test validation of path arguments"""
        # Test wrong number of arguments
        from jaf.exceptions import InvalidArgumentCountError

        with self.assertRaises(InvalidArgumentCountError):
            jaf_eval.eval(["@"], self.test_data)

    def test_path_argument_validation_2(self):
        from jaf.exceptions import InvalidArgumentCountError

        with self.assertRaises(InvalidArgumentCountError):
            jaf_eval.eval(["@", "user.name", "extra"], self.test_data)

    def test_path_argument_validation_3(self):
        # Test invalid path argument types
        from jaf.exceptions import InvalidQueryFormatError

        with self.assertRaises(InvalidQueryFormatError):
            jaf_eval.eval(["@", 123], self.test_data)

    def test_path_component_validation(self):
        """Test validation of path components after string conversion"""
        # This would be caught after string_to_path_ast conversion
        # if we had an invalid operation in the converted AST

        # Test manually constructed invalid AST to verify validation
        from jaf.exceptions import UnknownPathOperationError, InvalidQueryFormatError

        with self.assertRaises(UnknownPathOperationError):
            jaf_eval.eval(["@", [["invalid_op", "test"]]], self.test_data)

        with self.assertRaises(InvalidQueryFormatError):
            jaf_eval.eval(["@", ["not_a_list"]], self.test_data)

        with self.assertRaises(InvalidQueryFormatError):
            jaf_eval.eval(["@", [[]]], self.test_data)

    def test_empty_results_from_path_strings(self):
        """Test handling of paths that return no results"""
        # Non-existent key
        result = jaf_eval.eval(["@", "user.nonexistent"], self.test_data)
        self.assertEqual(result, [])

        # Out of bounds index
        result = jaf_eval.eval(["@", "items[10]"], self.test_data)
        self.assertEqual(result, [])

        # Slice with no matching elements
        result = jaf_eval.eval(["@", "items[10:20]"], self.test_data)
        self.assertEqual(result, [])

    def test_path_strings_with_special_characters(self):
        """Test path strings that include keys with special characters"""
        special_data = {
            "api_version": "1.0",
            "data-field": "test",
            "_private": "secret",
        }

        # These should work as regular keys
        result = jaf_eval.eval(["@", "api_version"], special_data)
        self.assertEqual(result, "1.0")

        # Keys with hyphens or underscores
        result = jaf_eval.eval(["@", "_private"], special_data)
        self.assertEqual(result, "secret")


if __name__ == "__main__":
    unittest.main()



"""
JAF DSL (Domain Specific Language) Parser using Lark

A clean grammar-based implementation of the DSL parser that converts
human-friendly filter expressions into JAF AST format.

Grammar Features:
- Operator precedence: or < and < not < comparison
- Function calls: length(@items), contains(@tags, "admin")
- Path expressions: @user.name, @items[0]
- Parentheses for grouping: (@age > 30 and @active) or @admin
- All standard literals: strings, numbers, booleans, null
"""

from lark import Lark, Transformer, v_args
from typing import List, Any, Union


# Define the grammar for our DSL
DSL_GRAMMAR = r"""
    ?start: or_expr

    ?or_expr: and_expr
            | or_expr "or" and_expr -> or_op

    ?and_expr: not_expr
             | and_expr "and" not_expr -> and_op

    ?not_expr: "not" atom -> not_op
             | atom

    ?atom: comparison
         | function_call
         | path
         | literal
         | "(" or_expr ")"

    comparison: atom comp_op atom

    function_call: IDENTIFIER "(" [arg_list] ")"

    path: "@" path_component ("." path_component)*

    arg_list: atom ("," atom)*

    path_component: IDENTIFIER
                  | "[" (NUMBER | STRING) "]" -> index_component

    comp_op: "==" -> eq
           | "!=" -> neq
           | ">"  -> gt
           | ">=" -> gte
           | "<"  -> lt
           | "<=" -> lte

    ?literal: STRING
            | NUMBER
            | "true"  -> true
            | "false" -> false
            | "null"  -> null
            | array

    array: "[" [atom ("," atom)*] "]"

    IDENTIFIER: /[a-zA-Z_][a-zA-Z0-9_]*/
    STRING: /"([^"\\]|\\.)*"/ | /'([^'\\]|\\.)*'/
    NUMBER: /-?\d+(\.\d+)?/

    %import common.WS
    %ignore WS
"""


@v_args(inline=True)  # Automatically unpack arguments
class DSLTransformer(Transformer):
    """Transform the parse tree into JAF AST format"""

    def start(self, expr):
        return expr

    # Logical operators
    def or_op(self, left, right):
        return ["or", left, right]

    def and_op(self, left, right):
        return ["and", left, right]

    def not_op(self, expr):
        return ["not", expr]

    # Comparison operators
    def comparison(self, left, op, right):
        return [op, left, right]

    def eq(self):
        return "eq?"

    def neq(self):
        return "neq?"

    def gt(self):
        return "gt?"

    def gte(self):
        return "gte?"

    def lt(self):
        return "lt?"

    def lte(self):
        return "lte?"

    # Function calls
    def function_call(self, name, args=None):
        func_name = str(name)
        args_list = args if args else []

        # Simple syntactic mapping: add '?' to predicates that need it
        # and handle hyphenation for multi-word functions
        hyphenated_predicates = {
            "contains": "contains?",
            "startswith": "starts-with?",
            "endswith": "ends-with?",
            "matches": "regex-match?",
            "regexmatch": "regex-match?",
            "closematch": "close-match?",
            "partialmatch": "partial-match?",
            "isstring": "is-string?",
            "isnumber": "is-number?",
            "isarray": "is-array?",
            "isobject": "is-object?",
            "isnull": "is-null?",
            "isempty": "is-empty?",
            "exists": "exists?",
        }

        # Apply mapping if it exists, otherwise pass through
        jaf_op = hyphenated_predicates.get(func_name, func_name)

        return [jaf_op, *args_list]

    def arg_list(self, *args):
        return list(args)

    # Path expressions
    def path(self, *components):
        path_parts = []
        for comp in components:
            if isinstance(comp, list):
                path_parts.extend(comp)
            else:
                path_parts.append(comp)
        return ["@", path_parts]

    def path_component(self, identifier):
        return [["key", str(identifier)]]

    def index_component(self, index):
        if isinstance(index, str):
            # String index like ["key"]
            return [["key", index]]
        else:
            # Numeric index like [0]
            return [["index", int(index)]]

    # Literals
    def STRING(self, s):
        # Remove quotes and handle escape sequences
        content = s[1:-1]  # Remove surrounding quotes
        content = content.replace('\\"', '"').replace("\\'", "'")
        content = content.replace("\\n", "\n").replace("\\t", "\t")
        content = content.replace("\\\\", "\\")
        return content

    def NUMBER(self, n):
        if "." in n:
            return float(n)
        return int(n)

    def IDENTIFIER(self, name):
        return str(name)

    def true(self):
        return True

    def false(self):
        return False

    def null(self):
        return None

    def array(self, *items):
        return list(items)


class DSLParser:
    """Parser for JAF's human-friendly DSL syntax using Lark"""

    def __init__(self):
        self.parser = Lark(DSL_GRAMMAR, parser="lalr", transformer=DSLTransformer())

    def parse(self, dsl_expression: str) -> List[Any]:
        """
        Parse a DSL expression into JAF AST format.

        Args:
            dsl_expression: Human-friendly filter expression

        Returns:
            JAF AST representation

        Raises:
            DSLSyntaxError: If the expression cannot be parsed
        """
        try:
            if not dsl_expression.strip():
                raise DSLSyntaxError("Empty expression")

            return self.parser.parse(dsl_expression)

        except Exception as e:
            # Convert Lark errors to our DSLSyntaxError
            raise DSLSyntaxError(f"Failed to parse expression: {e}") from e


class DSLSyntaxError(Exception):
    """Exception raised for DSL syntax errors"""

    pass


# Convenience function
def parse_dsl(expression: str) -> List[Any]:
    """Parse a DSL expression string into JAF AST format"""
    parser = DSLParser()
    return parser.parse(expression)



import unittest
from jaf.path_conversion import path_ast_to_string, string_to_path_ast
from jaf.path_exceptions import PathSyntaxError


class TestPathAstToString(unittest.TestCase):

    def test_empty_path(self):
        self.assertEqual(path_ast_to_string([]), "")

    def test_single_key(self):
        self.assertEqual(path_ast_to_string([["key", "name"]]), "name")

    def test_multiple_keys(self):
        self.assertEqual(
            path_ast_to_string(
                [["key", "user"], ["key", "address"], ["key", "street"]]
            ),
            "user.address.street",
        )

    def test_key_and_index(self):
        self.assertEqual(
            path_ast_to_string([["key", "items"], ["index", 0]]), "items[0]"
        )
        self.assertEqual(
            path_ast_to_string([["key", "items"], ["index", -1]]), "items[-1]"
        )

    def test_key_and_indices(self):
        self.assertEqual(
            path_ast_to_string([["key", "tags"], ["indices", [0, 2, 5]]]), "tags[0,2,5]"
        )
        self.assertEqual(
            path_ast_to_string([["key", "tags"], ["indices", [1]]]), "tags[1]"
        )

    def test_key_and_slice(self):
        self.assertEqual(
            path_ast_to_string([["key", "data"], ["slice", 1, 10, 2]]), "data[1:10:2]"
        )
        self.assertEqual(
            path_ast_to_string([["key", "data"], ["slice", None, 5, 1]]), "data[:5]"
        )
        self.assertEqual(
            path_ast_to_string([["key", "data"], ["slice", 2, None, 1]]), "data[2:]"
        )
        self.assertEqual(
            path_ast_to_string([["key", "data"], ["slice", 1, 5, None]]), "data[1:5]"
        )
        self.assertEqual(
            path_ast_to_string([["key", "data"], ["slice", None, None, 2]]), "data[::2]"
        )
        self.assertEqual(
            path_ast_to_string([["key", "data"], ["slice", None, 5, None]]), "data[:5]"
        )
        self.assertEqual(
            path_ast_to_string([["key", "data"], ["slice", 1, None, None]]), "data[1:]"
        )
        self.assertEqual(
            path_ast_to_string([["key", "data"], ["slice", None, None, None]]),
            "data[:]",
        )
        self.assertEqual(
            path_ast_to_string([["key", "data"], ["slice", None, None, 1]]), "data[:]"
        )

    def test_key_and_regex_key(self):
        self.assertEqual(
            path_ast_to_string([["key", "headers"], ["regex_key", "^X-"]]),
            "headers.~/^X-/",
        )

    def test_key_and_level_wildcard(self):
        self.assertEqual(
            path_ast_to_string([["key", "items"], ["wc_level"]]), "items[*]"
        )  # Corrected: removed dot

    def test_key_and_recursive_wildcard(self):
        self.assertEqual(
            path_ast_to_string([["key", "data"], ["wc_recursive"]]), "data.**"
        )

    def test_recursive_wildcard_at_start(self):
        self.assertEqual(path_ast_to_string([["wc_recursive"], ["key", "id"]]), "**.id")

    def test_level_wildcard_at_start(self):
        self.assertEqual(
            path_ast_to_string([["wc_level"], ["key", "name"]]), "[*].name"
        )

    def test_root_operator_string(self):
        self.assertEqual(path_ast_to_string([["root"]]), "#")
        self.assertEqual(path_ast_to_string([["root"], ["key", "config"]]), "#.config")
        self.assertEqual(
            path_ast_to_string([["key", "user"], ["root"], ["key", "config"]]),
            "user.#.config",
        )
        self.assertEqual(
            path_ast_to_string(
                [["key", "user"], ["index", 0], ["root"], ["key", "config"]]
            ),
            "user[0].#.config",
        )
        self.assertEqual(
            path_ast_to_string(
                [["key", "user"], ["slice", 0, 1, None], ["root"], ["key", "config"]]
            ),
            "user[0:1].#.config",
        )

    def test_complex_path(self):
        ast = [
            ["key", "users"],
            ["index", 0],
            ["key", "profile"],
            ["wc_level"],
            ["key", "detail"],
            ["slice", 1, None, None],
            ["root"],
            ["key", "metadata"],
            ["regex_key", "version_\\d+"],
        ]
        # Corrected: removed dot before [*]
        self.assertEqual(
            path_ast_to_string(ast),
            "users[0].profile[*].detail[1:].#.metadata.~/version_\\d+/",
        )

    def test_regex_key_at_start(self):
        self.assertEqual(
            path_ast_to_string([["regex_key", "error_\\d+"]]), "~/error_\\d+/"
        )

    def test_indices_at_start(self):
        self.assertEqual(path_ast_to_string([["indices", [0, 1]]]), "[0,1]")

    def test_slice_at_start(self):
        self.assertEqual(path_ast_to_string([["slice", 0, 1, None]]), "[0:1]")

    def test_invalid_component_format(self):
        # Test message now matches implementation
        with self.assertRaisesRegex(
            PathSyntaxError, "Invalid AST component format: expected a non-empty list."
        ):
            path_ast_to_string([["key", "user"], "not-a-list"])  # type: ignore
        with self.assertRaisesRegex(
            PathSyntaxError, "Invalid AST component format: expected a non-empty list."
        ):
            path_ast_to_string([["key", "user"], []])

    def test_invalid_regex_key(self):
        # Test message now matches implementation
        with self.assertRaisesRegex(
            PathSyntaxError, "'regex_key' operation expects a string pattern argument."
        ):
            path_ast_to_string([["regex_key", 123]])  # type: ignore

    def test_invalid_operation_arguments(self):
        # Test messages now match implementation
        with self.assertRaisesRegex(
            PathSyntaxError, "'key' operation expects a single string argument."
        ):
            path_ast_to_string([["key", 123]])  # type: ignore
        with self.assertRaisesRegex(
            PathSyntaxError, "'index' operation expects a single integer argument."
        ):
            path_ast_to_string([["index", "abc"]])  # type: ignore
        with self.assertRaisesRegex(
            PathSyntaxError,
            "'indices' operation expects a single list of integers argument.",
        ):
            path_ast_to_string([["indices", [1, "b"]]])  # type: ignore
        with self.assertRaisesRegex(
            PathSyntaxError,
            "'slice' operation expects 1 to 3 integer or None arguments",
        ):
            path_ast_to_string([["slice", "a", "b", "c"]])  # type: ignore
        with self.assertRaisesRegex(PathSyntaxError, "Slice step cannot be zero."):
            path_ast_to_string([["slice", None, None, 0]])
        with self.assertRaisesRegex(
            PathSyntaxError, "'wc_level' operation expects no arguments."
        ):
            path_ast_to_string([["wc_level", "arg"]])
        with self.assertRaisesRegex(
            PathSyntaxError, "'wc_recursive' operation expects no arguments."
        ):
            path_ast_to_string([["wc_recursive", "arg"]])
        with self.assertRaisesRegex(
            PathSyntaxError, "'root' operation expects no arguments."
        ):
            path_ast_to_string([["root", "arg"]])
        with self.assertRaisesRegex(
            PathSyntaxError, "Unknown JAF path component operation: 'unknown_op'"
        ):
            path_ast_to_string([["unknown_op", "arg"]])


class TestStringToPathAst(unittest.TestCase):

    def test_empty_string(self):
        self.assertEqual(string_to_path_ast(""), [])
        self.assertEqual(string_to_path_ast("   "), [])  # With whitespace

    def test_single_key(self):
        self.assertEqual(string_to_path_ast("name"), [["key", "name"]])

    def test_multiple_keys(self):
        self.assertEqual(
            string_to_path_ast("user.name"), [["key", "user"], ["key", "name"]]
        )
        self.assertEqual(
            string_to_path_ast("a.b.c"), [["key", "a"], ["key", "b"], ["key", "c"]]
        )

    def test_index(self):
        self.assertEqual(string_to_path_ast("[0]"), [["index", 0]])
        self.assertEqual(
            string_to_path_ast("items[0]"), [["key", "items"], ["index", 0]]
        )
        self.assertEqual(
            string_to_path_ast("items[-1].value"),
            [["key", "items"], ["index", -1], ["key", "value"]],
        )
        self.assertEqual(
            string_to_path_ast("items[  10  ]"), [["key", "items"], ["index", 10]]
        )  # Whitespace

    def test_indices(self):
        self.assertEqual(string_to_path_ast("[0,1,2]"), [["indices", [0, 1, 2]]])
        self.assertEqual(
            string_to_path_ast("items[0,1,-1]"),
            [["key", "items"], ["indices", [0, 1, -1]]],
        )
        self.assertEqual(
            string_to_path_ast("items[ 0 , 1 , 2 ]"),
            [["key", "items"], ["indices", [0, 1, 2]]],
        )  # Whitespace

    def test_slice_1(self):
        self.assertEqual(
            string_to_path_ast("[:]"), [["slice", 0, None]]
        )  # Step defaults to None in parser

    def test_slice_2(self):
        self.assertEqual(string_to_path_ast("[1:]"), [["slice", 1, None]])

    def test_slice_3(self):
        self.assertEqual(string_to_path_ast("[:5]"), [["slice", 0, 5]])

    def test_slice_4(self):
        self.assertEqual(string_to_path_ast("[1:5]"), [["slice", 1, 5]])

    def test_slice_5(self):
        self.assertEqual(string_to_path_ast("[1:10:2]"), [["slice", 1, 10, 2]])

    def test_slice_6(self):
        self.assertEqual(string_to_path_ast("[::2]"), [["slice", 0, None, 2]])

    def test_slice_7(self):
        self.assertEqual(string_to_path_ast("[10:1:-1]"), [["slice", 10, 1, -1]])

    def test_slice_8(self):
        self.assertEqual(
            string_to_path_ast("items[1:5:2]"), [["key", "items"], ["slice", 1, 5, 2]]
        )

    def test_slice_9(self):
        self.assertEqual(
            string_to_path_ast("items[ : ]"), [["key", "items"], ["slice", 0, None]]
        )  # Whitespace

    def test_slice_10(self):
        self.assertEqual(
            string_to_path_ast("items[1 : 5 : 2]"),
            [["key", "items"], ["slice", 1, 5, 2]],
        )  # Whitespace

    def test_wc_level(self):
        self.assertEqual(string_to_path_ast("[*]"), [["wc_level"]])
        self.assertEqual(
            string_to_path_ast("items[*]"), [["key", "items"], ["wc_level"]]
        )
        self.assertEqual(
            string_to_path_ast("items[*].name"),
            [["key", "items"], ["wc_level"], ["key", "name"]],
        )

    def test_wc_recursive(self):
        self.assertEqual(string_to_path_ast("**"), [["wc_recursive"]])
        self.assertEqual(
            string_to_path_ast("**.name"), [["wc_recursive"], ["key", "name"]]
        )
        self.assertEqual(
            string_to_path_ast("data.**"), [["key", "data"], ["wc_recursive"]]
        )  # At end
        self.assertEqual(
            string_to_path_ast("data.**.name"),
            [["key", "data"], ["wc_recursive"], ["key", "name"]],
        )

    def test_regex_key(self):
        self.assertEqual(
            string_to_path_ast("~/^item_\\d+$/"), [["regex_key", "^item_\\d+$"]]
        )
        self.assertEqual(
            string_to_path_ast("data.~/pattern/"),
            [["key", "data"], ["regex_key", "pattern"]],
        )
        self.assertEqual(
            string_to_path_ast("~/^a.b$/[0]"), [["regex_key", "^a.b$"], ["index", 0]]
        )

    def test_root(self):
        self.assertEqual(string_to_path_ast("#"), [["root"]])
        self.assertEqual(string_to_path_ast("#.config"), [["root"], ["key", "config"]])
        self.assertEqual(
            string_to_path_ast("user.#.name"),
            [["key", "user"], ["root"], ["key", "name"]],
        )

    def test_dot_handling(self):
        self.assertEqual(
            string_to_path_ast("key1.key2"), [["key", "key1"], ["key", "key2"]]
        )
        self.assertEqual(string_to_path_ast("key1[0]"), [["key", "key1"], ["index", 0]])
        # The current parser's eat_dot() will consume a dot if present after any component.
        self.assertEqual(
            string_to_path_ast("key1.[0]"), [["key", "key1"], ["index", 0]]
        )
        self.assertEqual(
            string_to_path_ast("key1.[*]"), [["key", "key1"], ["wc_level"]]
        )
        self.assertEqual(
            string_to_path_ast("key1.~/regex/"),
            [["key", "key1"], ["regex_key", "regex"]],
        )

    def test_complex_paths_1(self):
        self.assertEqual(
            string_to_path_ast("user.tags[*].name"),
            [["key", "user"], ["key", "tags"], ["wc_level"], ["key", "name"]],
        )

    def test_complex_paths_2(self):
        self.assertEqual(
            string_to_path_ast("**.items[0,1].value"),
            [["wc_recursive"], ["key", "items"], ["indices", [0, 1]], ["key", "value"]],
        )

    def test_complex_paths_3(self):
        self.assertEqual(
            string_to_path_ast(
                "data[1:5:2].*.id"
            ),  # Assuming `.*` is now parsed as wc_level
            [["key", "data"], ["slice", 1, 5, 2], ["wc_level"], ["key", "id"]],
        )  # Changed to wc_level

    def test_complex_paths_4(self):
        # To test `data[1:5:2].[*].id`
        self.assertEqual(
            string_to_path_ast("data[1:5:2].[*].id"),
            [["key", "data"], ["slice", 1, 5, 2], ["wc_level"], ["key", "id"]],
        )

    def test_complex_paths_5(self):
        self.assertEqual(
            string_to_path_ast("#.data[0].~/^field\\w+$/"),
            [["root"], ["key", "data"], ["index", 0], ["regex_key", "^field\\w+$"]],
        )

    def test_invalid_syntax_1(self):
        with self.assertRaisesRegex(PathSyntaxError, "Unterminated '\\['"):
            string_to_path_ast("items[0")

    def test_invalid_syntax_2(self):
        with self.assertRaisesRegex(PathSyntaxError, "Unterminated '\\['"):
            string_to_path_ast("items[0,1")  # Will be caught by find(']')

    def test_invalid_syntax_3(self):
        with self.assertRaisesRegex(
            PathSyntaxError, "Malformed slice"
        ):  # Or specific int parsing error
            string_to_path_ast("items[abc:]")

    def test_invalid_syntax_4(self):
        with self.assertRaisesRegex(PathSyntaxError, "Expected integer"):
            string_to_path_ast("items[abc]")

    def test_invalid_syntax_5(self):
        with self.assertRaisesRegex(PathSyntaxError, "Expected integer"):
            string_to_path_ast("items[0,abc,1]")

    def test_invalid_syntax_6(self):
        with self.assertRaisesRegex(PathSyntaxError, "Unterminated regex key"):
            string_to_path_ast("~/pattern")

    def test_invalid_syntax_7(self):
        with self.assertRaisesRegex(PathSyntaxError, "Unexpected token"):
            string_to_path_ast("item!")

    def test_invalid_syntax_8(self):
        with self.assertRaisesRegex(
            PathSyntaxError, "Unexpected token"
        ):  # Dot at start
            string_to_path_ast(".item")

    def test_slice_parsing_edge_cases_1(self):
        x = string_to_path_ast("data[1:]")
        self.assertEqual(x, [["key", "data"], ["slice", 1, None]])

    def test_slice_parsing_edge_cases_2(self):
        self.assertEqual(
            string_to_path_ast("data[:5:]"), [["key", "data"], ["slice", 0, 5]]
        )  # step is None

    def test_slice_parsing_edge_cases_3(self):
        self.assertEqual(
            string_to_path_ast("data[::]"), [["key", "data"], ["slice", 0, None]]
        )  # step is None

    def test_slice_parsing_edge_cases_4(self):
        # Test for slice step zero - current string_to_path_ast does not validate this, path_ast_to_string does.
        # The parser should accept it, validation can be a separate step if needed.
        self.assertEqual(string_to_path_ast("[::0]"), [["slice", 0, None, 0]])


if __name__ == "__main__":
    unittest.main()



"""
Lazy operation loaders that wrap other streams with additional functionality.

These loaders implement lazy operations like take, skip, etc. as composable
stream transformations.
"""

from typing import Generator, Dict, Any, Optional, List
import itertools
from .exceptions import QueryError


def stream_take(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Take the first n items from an inner source.

    Args:
        source: Dict with:
            - inner_source: Source to take from
            - n: Number of items to take
    """
    inner_source = source.get("inner_source")
    n = source.get("n", 10)

    if not inner_source:
        raise ValueError("Take source missing 'inner_source'")

    count = 0
    for item in loader.stream(inner_source):
        if count >= n:
            break
        yield item
        count += 1


def stream_skip(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Skip the first n items from an inner source.

    Args:
        source: Dict with:
            - inner_source: Source to skip from
            - n: Number of items to skip
    """
    inner_source = source.get("inner_source")
    n = source.get("n", 0)

    if not inner_source:
        raise ValueError("Skip source missing 'inner_source'")

    for i, item in enumerate(loader.stream(inner_source)):
        if i >= n:
            yield item


def stream_slice(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Slice items from an inner source.

    Args:
        source: Dict with:
            - inner_source: Source to slice from
            - start: Starting index
            - stop: Stopping index (exclusive), None for no limit
            - step: Step size (default: 1)
    """
    inner_source = source.get("inner_source")
    start = source.get("start", 0)
    stop = source.get("stop")
    step = source.get("step", 1)

    if not inner_source:
        raise ValueError("Slice source missing 'inner_source'")

    for item in itertools.islice(loader.stream(inner_source), start, stop, step):
        yield item


def stream_filter(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Filter items from an inner source using a JAF query.

    Args:
        source: Dict with:
            - inner_source: Source to filter
            - query: JAF query to use as predicate
    """
    from .jaf_eval import jaf_eval

    inner_source = source.get("inner_source")
    query = source.get("query")

    if not inner_source:
        raise ValueError("Filter source missing 'inner_source'")
    if not query:
        raise ValueError("Filter source missing 'query'")

    for item in loader.stream(inner_source):
        try:
            result = jaf_eval.eval(query, item)
            if isinstance(result, bool) and result:
                yield item
        except (KeyError, AttributeError, TypeError, IndexError):
            # Expected "item doesn't match" errors - skip the item
            # These happen when paths don't exist, types don't match, etc.
            pass
        except QueryError:
            # Query errors (unknown operators, invalid syntax) should fail fast
            raise
        # Let other unexpected exceptions propagate


def stream_take_while(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Take items while a predicate query is true.

    Args:
        source: Dict with:
            - inner_source: Source to take from
            - query: JAF query to use as predicate
    """
    from .jaf_eval import jaf_eval

    inner_source = source.get("inner_source")
    query = source.get("query")

    if not inner_source:
        raise ValueError("Take-while source missing 'inner_source'")
    if not query:
        raise ValueError("Take-while source missing 'query'")

    for item in loader.stream(inner_source):
        try:
            result = jaf_eval.eval(query, item)
            if isinstance(result, bool) and result:
                yield item
            else:
                break
        except Exception:
            break


def stream_skip_while(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Skip items while a predicate query is true.

    Args:
        source: Dict with:
            - inner_source: Source to skip from
            - query: JAF query to use as predicate
    """
    from .jaf_eval import jaf_eval

    inner_source = source.get("inner_source")
    query = source.get("query")

    if not inner_source:
        raise ValueError("Skip-while source missing 'inner_source'")
    if not query:
        raise ValueError("Skip-while source missing 'query'")

    skipping = True
    for item in loader.stream(inner_source):
        if skipping:
            try:
                result = jaf_eval.eval(query, item)
                if isinstance(result, bool) and result:
                    continue
                else:
                    skipping = False
            except Exception:
                skipping = False

        if not skipping:
            yield item


def stream_batch(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Batch items from an inner source into groups.

    Args:
        source: Dict with:
            - inner_source: Source to batch
            - size: Batch size
    """
    inner_source = source.get("inner_source")
    size = source.get("size", 10)

    if not inner_source:
        raise ValueError("Batch source missing 'inner_source'")

    batch = []
    for item in loader.stream(inner_source):
        batch.append(item)
        if len(batch) >= size:
            yield batch
            batch = []

    # Yield any remaining items
    if batch:
        yield batch


def stream_enumerate(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Enumerate items with an index.

    Args:
        source: Dict with:
            - inner_source: Source to enumerate
            - start: Starting index (default: 0)
            - as_dict: If True, yield {"index": i, "value": item}, else yield [i, item]
    """
    inner_source = source.get("inner_source")
    start = source.get("start", 0)
    as_dict = source.get("as_dict", True)

    if not inner_source:
        raise ValueError("Enumerate source missing 'inner_source'")

    for i, item in enumerate(loader.stream(inner_source), start=start):
        if as_dict:
            yield {"index": i, "value": item}
        else:
            yield [i, item]


def stream_map(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Transform items using a JAF query/expression.

    Args:
        source: Dict with:
            - inner_source: Source to map
            - expression: JAF expression to evaluate for each item
    """
    from .jaf_eval import jaf_eval

    inner_source = source.get("inner_source")
    expression = source.get("expression")

    if not inner_source:
        raise ValueError("Map source missing 'inner_source'")
    if not expression:
        raise ValueError("Map source missing 'expression'")

    for item in loader.stream(inner_source):
        try:
            result = jaf_eval.eval(expression, item)
            yield result
        except Exception as e:
            # Could optionally yield None or skip
            yield None


def stream_chain(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Chain multiple sources sequentially.

    Args:
        source: Dict with:
            - sources: List of sources to chain
    """
    sources = source.get("sources", [])

    for src in sources:
        yield from loader.stream(src)


def stream_join(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Join two streams based on a key expression.

    Args:
        source: Dict with:
            - left: Left source
            - right: Right source
            - on: JAF expression to extract join key
            - how: Join type ("inner", "left", "right", "outer")
    """
    from .jaf_eval import jaf_eval

    left_source = source.get("left")
    right_source = source.get("right")
    on_expr = source.get("on")
    how = source.get("how", "inner")

    if not left_source or not right_source:
        raise ValueError("Join source missing 'left' or 'right'")
    if not on_expr:
        raise ValueError("Join source missing 'on' expression")

    # For now, implement a simple in-memory hash join
    # TODO: For large streams, consider sort-merge join or nested loop with buffering

    # Build index from right stream
    right_index = {}
    for item in loader.stream(right_source):
        try:
            key = jaf_eval.eval(on_expr, item)
            if key not in right_index:
                right_index[key] = []
            right_index[key].append(item)
        except Exception:
            pass

    # Stream left and join
    for left_item in loader.stream(left_source):
        try:
            key = jaf_eval.eval(on_expr, left_item)
            right_matches = right_index.get(key, [])

            if right_matches:
                # Inner join - output combinations
                for right_item in right_matches:
                    yield {"left": left_item, "right": right_item}
            elif how in ("left", "outer"):
                # Left join - output with null right
                yield {"left": left_item, "right": None}
        except Exception:
            if how in ("left", "outer"):
                yield {"left": left_item, "right": None}

    # For right/outer joins, emit unmatched right items
    if how in ("right", "outer"):
        # Need to track which right items were matched
        # This simple implementation doesn't do that efficiently
        # TODO: Improve join implementation
        pass


def stream_groupby(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Group items from a stream by a key expression.

    Args:
        source: Dict with:
            - inner_source: Source to group
            - key: JAF expression to extract grouping key
            - aggregate: Optional aggregation operations (dict mapping field names to ops)

    Example:
        {
            "type": "groupby",
            "inner_source": {...},
            "key": "@category",
            "aggregate": {
                "count": ["count"],
                "total_price": ["sum", "@price"],
                "avg_price": ["mean", "@price"],
                "max_price": ["max", "@price"]
            }
        }
    """
    from .jaf_eval import jaf_eval
    import statistics

    inner_source = source.get("inner_source")
    key_expr = source.get("key")
    aggregate = source.get("aggregate", {})

    if not inner_source:
        raise ValueError("Groupby source missing 'inner_source'")
    if not key_expr:
        raise ValueError("Groupby source missing 'key' expression")

    # Collect all items into groups
    groups = {}
    for item in loader.stream(inner_source):
        try:
            key = jaf_eval.eval(key_expr, item)
            # Convert unhashable types to strings for grouping
            if isinstance(key, (list, dict)):
                key = str(key)
            if key not in groups:
                groups[key] = []
            groups[key].append(item)
        except Exception:
            # Items that can't be grouped go into None group
            if None not in groups:
                groups[None] = []
            groups[None].append(item)

    # Yield groups with optional aggregation
    for key, items in groups.items():
        result = {"key": key, "items": items, "count": len(items)}

        # Apply aggregations
        for field_name, agg_spec in aggregate.items():
            if not isinstance(agg_spec, list) or len(agg_spec) == 0:
                continue

            op = agg_spec[0]
            value_expr = agg_spec[1] if len(agg_spec) > 1 else "@"

            # Extract values for aggregation
            values = []
            for item in items:
                try:
                    val = jaf_eval.eval(value_expr, item)
                    if val is not None:
                        values.append(val)
                except Exception:
                    pass

            # Apply aggregation operation
            if op == "count":
                result[field_name] = len(items)
            elif op == "sum" and values:
                result[field_name] = sum(values)
            elif op == "mean" and values:
                result[field_name] = statistics.mean(values)
            elif op == "median" and values:
                result[field_name] = statistics.median(values)
            elif op == "stddev" and len(values) > 1:
                result[field_name] = statistics.stdev(values)
            elif op == "variance" and len(values) > 1:
                result[field_name] = statistics.variance(values)
            elif op == "min" and values:
                result[field_name] = min(values)
            elif op == "max" and values:
                result[field_name] = max(values)
            elif op == "first" and items:
                result[field_name] = jaf_eval.eval(value_expr, items[0])
            elif op == "last" and items:
                result[field_name] = jaf_eval.eval(value_expr, items[-1])
            else:
                result[field_name] = None

        yield result


def stream_product(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Compute the Cartesian product of two streams.

    Args:
        source: Dict with:
            - left: Left source
            - right: Right source
            - limit: Optional limit on output size (since product can be very large)
    """
    left_source = source.get("left")
    right_source = source.get("right")
    limit = source.get("limit")

    if not left_source or not right_source:
        raise ValueError("Product source missing 'left' or 'right'")

    # Collect right stream into memory (needed for product)
    right_items = list(loader.stream(right_source))

    if not right_items:
        return  # Empty product

    count = 0
    for left_item in loader.stream(left_source):
        for right_item in right_items:
            if limit and count >= limit:
                return

            yield {"left": left_item, "right": right_item}
            count += 1


def stream_distinct(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Remove duplicate items from a stream.

    Args:
        source: Dict with:
            - inner_source: Source to deduplicate
            - key: Optional JAF expression to extract uniqueness key
                  If not provided, uses the entire item
    """
    from .jaf_eval import jaf_eval
    import json

    inner_source = source.get("inner_source")
    key_expr = source.get("key")

    if not inner_source:
        raise ValueError("Distinct source missing 'inner_source'")

    seen = set()

    for item in loader.stream(inner_source):
        try:
            if key_expr:
                # Use key expression for uniqueness
                key = jaf_eval.eval(key_expr, item)
            else:
                # Use entire item
                key = item

            # Convert unhashable types to strings
            if isinstance(key, (list, dict)):
                key = json.dumps(key, sort_keys=True)

            if key not in seen:
                seen.add(key)
                yield item
        except Exception:
            # On error, consider item unique
            yield item


def stream_project(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Project specific fields from items (like SQL SELECT).

    Args:
        source: Dict with:
            - inner_source: Source to project from
            - fields: Dict mapping output field names to JAF expressions

    Example:
        {
            "type": "project",
            "inner_source": {...},
            "fields": {
                "name": "@name",
                "age": "@age",
                "full_address": ["join", ", ", ["@address.street", "@address.city"]]
            }
        }
    """
    from .jaf_eval import jaf_eval

    inner_source = source.get("inner_source")
    fields = source.get("fields", {})

    if not inner_source:
        raise ValueError("Project source missing 'inner_source'")

    for item in loader.stream(inner_source):
        result = {}
        for field_name, expression in fields.items():
            try:
                result[field_name] = jaf_eval.eval(expression, item)
            except Exception:
                result[field_name] = None
        yield result


def stream_union(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Union multiple streams (concatenate with optional deduplication).

    Args:
        source: Dict with:
            - sources: List of sources to union
            - distinct: If True, remove duplicates (default: False)
    """
    import json

    sources = source.get("sources", [])
    distinct = source.get("distinct", False)

    if distinct:
        seen = set()

    for src in sources:
        for item in loader.stream(src):
            if distinct:
                # Create hashable key
                key = item
                if isinstance(key, (list, dict)):
                    key = json.dumps(key, sort_keys=True)

                if key in seen:
                    continue
                seen.add(key)

            yield item


def stream_intersect(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Intersect two streams (items that appear in both).

    Args:
        source: Dict with:
            - left: First source
            - right: Second source
            - key: Optional JAF expression for comparison key
    """
    from .jaf_eval import jaf_eval
    import json

    left_source = source.get("left")
    right_source = source.get("right")
    key_expr = source.get("key")

    if not left_source or not right_source:
        raise ValueError("Intersect source missing 'left' or 'right'")

    # Collect right stream into set
    right_set = set()
    for item in loader.stream(right_source):
        try:
            if key_expr:
                key = jaf_eval.eval(key_expr, item)
            else:
                key = item

            if isinstance(key, (list, dict)):
                key = json.dumps(key, sort_keys=True)

            right_set.add(key)
        except Exception:
            pass

    # Stream left and check membership
    seen = set()
    for item in loader.stream(left_source):
        try:
            if key_expr:
                key = jaf_eval.eval(key_expr, item)
            else:
                key = item

            if isinstance(key, (list, dict)):
                key = json.dumps(key, sort_keys=True)

            if key in right_set and key not in seen:
                seen.add(key)
                yield item
        except Exception:
            pass


def stream_except(
    loader: "StreamingLoader", source: Dict[str, Any]
) -> Generator[Any, None, None]:
    """
    Except/difference of two streams (items in left but not in right).

    Args:
        source: Dict with:
            - left: First source
            - right: Second source to subtract
            - key: Optional JAF expression for comparison key
    """
    from .jaf_eval import jaf_eval
    import json

    left_source = source.get("left")
    right_source = source.get("right")
    key_expr = source.get("key")

    if not left_source or not right_source:
        raise ValueError("Except source missing 'left' or 'right'")

    # Collect right stream into set
    right_set = set()
    for item in loader.stream(right_source):
        try:
            if key_expr:
                key = jaf_eval.eval(key_expr, item)
            else:
                key = item

            if isinstance(key, (list, dict)):
                key = json.dumps(key, sort_keys=True)

            right_set.add(key)
        except Exception:
            pass

    # Stream left and check non-membership
    for item in loader.stream(left_source):
        try:
            if key_expr:
                key = jaf_eval.eval(key_expr, item)
            else:
                key = item

            if isinstance(key, (list, dict)):
                key = json.dumps(key, sort_keys=True)

            if key not in right_set:
                yield item
        except Exception:
            # On error, include the item
            yield item


# Register the lazy operation loaders
def register_lazy_ops_loaders(loader: "StreamingLoader"):
    """Register all lazy operation loaders with the streaming loader."""
    loader.register("take", stream_take)
    loader.register("skip", stream_skip)
    loader.register("slice", stream_slice)
    loader.register("filter", stream_filter)
    loader.register("take_while", stream_take_while)
    loader.register("skip_while", stream_skip_while)
    loader.register("batch", stream_batch)
    loader.register("enumerate", stream_enumerate)
    loader.register("map", stream_map)
    loader.register("chain", stream_chain)
    loader.register("join", stream_join)
    loader.register("groupby", stream_groupby)
    loader.register("product", stream_product)
    loader.register("distinct", stream_distinct)
    loader.register("project", stream_project)
    loader.register("union", stream_union)
    loader.register("intersect", stream_intersect)
    loader.register("except", stream_except)



"""
JAF Custom Exception Classes
============================

This module defines custom exceptions for better error handling and debugging.
Exceptions are organized into categories based on when they occur and how they
should be handled.
"""


class JAFError(Exception):
    """Base exception for all JAF errors"""

    pass


# Query/Syntax Errors (should fail fast during evaluation)
class QueryError(JAFError):
    """Base for all query-related errors that indicate invalid queries"""

    pass


class UnknownOperatorError(QueryError):
    """Raised when query contains unknown operator"""

    def __init__(self, operator):
        super().__init__(f"Unknown operator: {operator}")
        self.operator = operator


class InvalidArgumentCountError(QueryError):
    """Raised when operator gets wrong number of arguments"""

    def __init__(self, operator, expected, got):
        if expected == -1:
            # Variadic operator
            super().__init__(f"'{operator}' expects variable arguments, got {got}")
        else:
            super().__init__(f"'{operator}' expects {expected} arguments, got {got}")
        self.operator = operator
        self.expected = expected
        self.got = got


class InvalidQueryFormatError(QueryError):
    """Raised when query has invalid format"""

    pass


# Path Errors
class PathError(JAFError):
    """Base for path-related errors"""

    pass


class PathSyntaxError(PathError):
    """Raised when path has syntax errors"""

    def __init__(self, message, path_segment=None):
        super().__init__(message)
        self.path_segment = path_segment


class UnknownPathOperationError(PathError):
    """Raised when path contains unknown operation"""

    def __init__(self, operation):
        super().__init__(f"Unknown path operation: {operation}")
        self.operation = operation


# Evaluation Errors (might be item-specific, safe to skip)
class EvaluationError(JAFError):
    """Base for errors during evaluation that might be item-specific"""

    pass


class PathNotFoundError(EvaluationError):
    """Raised when path doesn't exist in data (like KeyError)"""

    pass


class TypeMismatchError(EvaluationError):
    """Raised when types don't match expected (like TypeError)"""

    pass


class IndexOutOfBoundsError(EvaluationError):
    """Raised when array index is out of bounds"""

    pass
